// Copyright 2023 xwst <xwst@gmx.net>

/*
This file is part of sinaps.

sinaps is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

sinaps is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with sinaps. If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef SINAPS_BACKEND_H
#define SINAPS_BACKEND_H

#include "backend/backend.h"

#include <sstream>

#include <giomm.h>

#include "backend/parser.h"
#include "backend/util.h"

namespace {

Glib::ustring usage_string() {
  return "Usage:\nDerive password using a line of the form\n"
         "> context [length(int) [counter(int) [alphabet]]]\n"
         "\nor use one of the following commands:\n"
         "\talphabet\n"
         "\tclear\n"
         "\texit | quit\n"
         "\thelp | usage\n"
         "\tversion\n";
}


void replace_substr(
  Glib::ustring& s,
  const Glib::ustring& from,
  const Glib::ustring& to
) {
  auto location = s.find(from);
  if (location != Glib::ustring::npos)
    s.replace(location, from.size(), to);
}


void replace_alphabet_abbreviations(Glib::ustring& s) {
  for (const auto& [key, value] : sinaps::alphabet_abbreviations)
    replace_substr(s, key, value);
}


}

/**
 * \namespace sinaps
 * \brief Project wide namespace. The documentation contains a description of
 *        the password derivation process.
 *
 * \section Overview
 * In conventional password managers, passwords are stored in a database and
 * encrypted using a master password. Stateless password managers like sinaps
 * on the other hand derive passwords using key derivation functions (KDFs)
 * based on two pieces of information:
 * - a main password,
 * - context information such as a website, service name or user name, counter
 *   and encoding.
 *
 * Instead of calculating the passwords directly, sinaps transforms the main
 * password into an intermediate passphrase using a costly KDF to mitigate weak
 * user inputs but allow a quick derivation of passwords.
 * The passwords are then derived from the intermediate passphrase.
 * The following naming scheme applies:
 * | **Name**                 | **Description**                                |
 * |--------------------------|------------------------------------------------|
 * | Pin                      | The main secret entered by the user.           |
 * | \ref Passphrase          | Auxiliary secret derived from pin using a KDF. |
 * | \ref Password            | Final output to be used for services.          |
 * | \ref Context             | Service specific information. (See below!)     |
 * | \ref Collection          | Set of contexts. (See below!)                  |
 *
 *
 * sinaps employs the [libsodium](https://doc.libsodium.org/) library for all
 * cryptographic computations.
 *
 * \section Passphrase
 * The \ref SecureMemory::Passphrase "passphrase" is a 256 bit pseudo random
 * number derived from the pin using the Argon2 hash function. Argon2 also
 * accepts two additional parameters specifying the utilized memory and
 * computational expense; changing these parameters will also alter the result.
 * Argon2 is used to mitigate the weakness of the user pin to some extend, but
 * the user is still expected to choose a strong pin. The passphrase is assumed
 * to be cryptographically secure, which is only true, if the pin is not too
 * weak.
 *
 * \section Password
 * \subsection Calculation
 * Passwords are the final secrets derived by sinaps that are meant to be used
 * outside of sinaps. They are derived from the passphrase using a keyed hash
 * of the context name, where the passphrase is used as the key. The applied
 * hash function is BLAKE2b, which is sufficient provided that the passphrase is
 * cryptographically secure.
 * The password derivation consists two steps:
 * 1. A 256 bit pseudo random seed is obtained by hashing the Context name as
 *    described above. This is required to convert the arbitrary context string
 *    to a byte sequence of fixed size.
 * 2. The IETF variant of ChaCha20 is used to create a stream of pseudo random
 *    bytes by using the seed as a key. Each 64 byte block in the stream is
 *    converted to a single password character.
 *
 * ChaCha20 also depends on a nonce; the context counter is converted to an
 * ASCII hex number and used as the nonce which allows to obtain a new password
 * for a given service by just increasing the counter.
 *
 * \subsection Limits
 * \subsubsection Length
 * The IETF variant of ChaCha20 supports the generation of up to
 * \f$2^{32} = 4\,294\,967\,296\f$ blocks of 64 bytes which determines the
 * maximal password length. However, in order to avoid errors when trying to
 * allocate memory and keep the program responsive, the actual password length
 * is restricted to \f$2^{16} = 65\,536\f$ (see \ref sinaps::Context). If large
 * amounts of deterministically derived data are required, sinaps can be used to
 * calculate a seed for a third party tool generating pseudo random numbers.
 *
 * \subsubsection Counter
 * The counter used in sinaps is passed to the ChaCha20-algorithm as a nonce
 * guaranteeing that altering the counter will produce a new password. The size
 * of the nonce in IETF-ChaCha20 is 12 bytes, which corresponds to at most
 * \f$2^{8\cdot 12} = 2^{96}\f$ different passwords for a given context name.
 * The current sinaps implementation is more restritive and allows to use only
 * unsigned 64-bit numbers for the counter giving a maximum of \f$2^{64}\f$
 * different passwords per pin and context name.
 *
 * \section Context
 * A Context object consist of the following fields:
 * | **Field** | **Description**                                              |
 * |-----------|--------------------------------------------------------------|
 * | Name      | Service/website/user name or any other unique identifyer.    |
 * | Length    | Length (number of characters) of the password to be derived. |
 * | Counter   | Used to obtain a new password without changing other fields. |
 * | Alphabet  | Set of characters that may appear in the password.           |
 * | Skip      | If \f$\neq 0\f$, the first `Skip` characters will be skipped.|
 *
 * \section Collection
 * A set of Context objects that represents a configuration file or a history of
 * user inputs in an interactive session. The collection also stores default
 * values for parameters like the alphabet or the length of a password.
 */
namespace sinaps {

/**
 * @var Backend::signal_argon2_parameters_missing
 * @brief This signal is emitted to inform the frontend which Argon2 have not
 *        been specified yet and need to be queried from the user.
 *
 * This signal sends three parameters:
 * @arg <b>Glib::ustring</b>
 *  This parameter indicates whether the user has to enter a pin. If the string
 *  is empty, a pin is required; if not, the string will contain information
 *  about the existing pin and can be displayed without disclosing information.
 * @arg <b>unsigned long long</b>
 *  This parameter corresponds to the second argument of the
 *  @ref SecureMemory::Passphrase::Passphrase(
 *         Vector_uptr<T>,
 *         unsigned long long,
 *         size_t
 *       ) "Passphrase constructor". A zero value means that an input is
 *  required; a non-zero value represents the current value and can be presented
 *  to the user.
 * @arg <b>size_t</b>
 *  This parameter corresponds to the third argument of the
 *  @ref SecureMemory::Passphrase::Passphrase(
 *         Vector_uptr<T>,
 *         unsigned long long,
 *         size_t
 *       ) "Passphrase constructor". A zero value means that an input is
 *  required; a non-zero value represents the current value and can be presented
 *  to the user.
 */


/**
 * @var Backend::signal_start_frontend
 * @brief This signal indicates that the frontend should be started. It is
 *        emitted when the command line arguments have been parsed.
 */


/**
 * @var Backend::signal_start_input_loop
 * @brief This signal is emitted as soon as the Passphrase is set up and
 *        indicates that the frontend can start accepting user input for
 *        password derivations.
 */


/**
 * @var Backend::signal_display_password
 *
 * The password is passed as a reference to a unique pointer as libsigc++
 * does not support move semantics.
 *
 * @warning This signal must be connected to only one slot, as the frontend will
 *          invalidate the unique pointer. A second connection will lead to
 *          undefined behaviour.
 * @sa FrontendInterface::display_password
 */


/**
 * @var Backend::signal_display_password
 *
 * The password is passed as a reference to a unique pointer as libsigc++
 * does not support move semantics.
 *
 * @warning This signal must be connected to only one slot, as the frontend will
 *          invalidate the unique pointer. A second connection will lead to
 *          undefined behaviour.
 * @sa FrontendInterface::password_to_clipboard
 */


Glib::ustring Backend::get_version_string() {
  return "0.8";
}

/**
  * @brief Backend constructor.
  */
Backend::Backend() {
  this->opslimit = 0;
  this->memlimit = 0;
}


/**
 * @brief Check if all parameters required for the Passphrase are set.
 */
bool Backend::argon2_parameters_ready() {
  return (this->opslimit > 0)
     and (this->memlimit > 0)
     and (this->pin != nullptr);
}


void Backend::initialize_passphrase_if_ready() {
  if (this->argon2_parameters_ready()) {
    this->p = std::make_unique<SecureMemory::Passphrase<gunichar>>(
      std::move(this->pin),
      this->opslimit,
      this->memlimit
    );
    this->signal_start_input_loop.emit();
  }
}


/**
 * @brief Instruct the frontend to ask for Argon2 parameters.
 */
void Backend::on_frontend_ready() {
  // Emit signal to prompt user for Argon2 parameters.
  this->signal_argon2_parameters_missing("", this->opslimit, this->memlimit);
}


void Backend::set_opslimit(unsigned long long opslimit) {
  if (opslimit > 0) {
    this->opslimit = opslimit;
    this->initialize_passphrase_if_ready();
  }
}


/**
 * @brief Store the memory to be used for the Passphrase derivation at a later
 *        point in time.
 * @param memlimit The amount of memory to be use in mebibytes.
 */
void Backend::set_memlimit(size_t memlimit) {
  if (memlimit > 0) {
    // libsodium expects bytes instead of mebibytes:
    this->memlimit = memlimit * 1048576;
    this->initialize_passphrase_if_ready();
  }
}


void Backend::set_pin(GCV_uptr& pin) {
  this->pin = std::move(pin);
  this->initialize_passphrase_if_ready();
}


/**
  * @brief This function should be called (via signals) when the frontend is
  *        attempted to be closed.
  *
  * It could be used to ask for certain user input a last time before the
  * frontend is shutdown.
  */
void Backend::on_frontend_closed() {
  // Handle close request of frontend if necessary.
  this->p.reset();
}


/**
 * @brief Handle special one word commands like "help" or "exit".
 * @param command A string containing a single word.
 * @return true if the command was processed or false if the passed word is not
 *         a recognized command.
 */
bool Backend::handle_special_command(const Glib::ustring& command) {
  if (command == "exit" || command == "quit") {
    this->p.reset();
    this->signal_exit.emit();
    return true;
  } else if (command == "clear") {
    this->signal_clear_output.emit();
    return true;
  } else if (command == "version") {
    this->signal_display_info(this->get_version_string());
    return true;
  } else if (command == "usage" || command == "help" || command == "?") {
    this->signal_display_info.emit(usage_string());
    return true;
  } else if (command == "alphabet") {
    this->signal_display_info.emit("Alphabet substitutions:");
    for (auto [key, value] : sinaps::alphabet_abbreviations)
      this->signal_display_info.emit(Glib::ustring::compose(
        "%1%2-->     %3", key, Glib::ustring(15 - key.length(), ' '), value
      ));
    return true;
  }
  return false;
}


/**
 * @brief Calculate the password for the given input and emit the corresonding
 *        signal.
 *
 * The line is parsed to determine the provided input parameters and the
 * collection is searched for a match to obtain missing parameters from the
 * configuration file (if any).
 *
 * @param ctx A vector of the whitespace-separated pieces of the entered line.
 */
void Backend::handle_context_input(std::vector<Glib::ustring>& ctx) {
  auto input = Config::ConfigParser::parse_context(ctx);
  auto config = (this->collection == nullptr)
                ? input : this->collection->get_child_params(ctx[0]);
  // When no configuration file was specified, use 'input' once again to avoid
  // branching in the following.

  // Choose provided input parameter if it exists. If not, choose value from
  // configuration file. Fall back to default values.
  auto choose = [](auto& opt1, auto& opt2, auto default_value) {
    opt1 = opt1.value_or(opt2.value_or(default_value));
  };
  choose(input.counter,  config.counter,  default_counter);
  choose(input.length,   config.length,   default_length);
  choose(input.alphabet, config.alphabet, default_alphabet);
  choose(input.skip,     config.skip,     default_skip);

  replace_alphabet_abbreviations(input.alphabet.value());

  auto name = ustr_to_UCS4_vec(input.name);
  auto alphabet = ustr_to_UCS4_vec(input.alphabet.value());
  auto pw = this->p->custom(
    name,
    input.counter.value(),
    input.length.value(),
    alphabet,
    input.skip.value()
  );
  this->signal_display_password.emit(pw);
}


/**
 * @brief This function handles user input by parsing the entered string and
 *        performing the required actions.
 */
void Backend::on_line_entered(const Glib::ustring& line) {
  // Parse input line
  auto cmd = Glib::Regex::split_simple("\\s+", line);
  std::erase(cmd, "");

  if(cmd.size() == 0) return; // only white spaces
  if(cmd.size() == 1)// one-word commands
    if (this->handle_special_command(cmd[0]))
      return;

  try {
    this->handle_context_input(cmd);
  } catch(const std::domain_error& e) {
    this->signal_display_info.emit(e.what());
  }
}


/**
 * @brief This function updates the list of possible completions by emitting
 *        signal_new_input_completions.
 */
void Backend::on_input_buffer_changed(const Glib::ustring& buffer) {
  // Update completions
  if (this->collection != nullptr)
    this->signal_new_input_completions.emit(
      this->collection->find_children(buffer, true)
    );
}


/**
 * @brief Parse the command line arguments that are passed when starting the
 *        application.
 */
int Backend::parse_commandline_arguments(
  const Glib::RefPtr<Gio::ApplicationCommandLine>& cl
) {
  int argc;
  char** argv = cl->get_arguments(argc);

  Glib::OptionContext oc("[FILE_NAME]");
  oc.set_summary("Derive secrets from a single passphrase");

  Glib::OptionGroup og("", "", "");
  // Arguments are irrelevant because this option group is the only one.

  bool version = false;
  Glib::OptionEntry oe_version = Glib::OptionEntry();
  oe_version.set_long_name("version");
  oe_version.set_short_name('v');
  oe_version.set_description("Print version information and quit");
  og.add_entry(oe_version, version);

  Glib::OptionEntry oe_ops = Glib::OptionEntry();
  oe_ops.set_long_name("iter");
  oe_ops.set_short_name('i');
  oe_ops.set_arg_description("ITERATIONS");
  oe_ops.set_description("Number of iterations");
  auto oh = [&](const Glib::ustring& name, const Glib::ustring& value, bool) {
    std::ignore = name;
    return Config::parse_input_data(value, this->opslimit);
  };
  og.add_entry(oe_ops, oh);

  Glib::OptionEntry oe_mem = Glib::OptionEntry();
  oe_mem.set_long_name("memory");
  oe_mem.set_short_name('m');
  oe_mem.set_arg_description("MEMORY");
  oe_mem.set_description("Used memory in mebibytes");
  auto mh = [&](const Glib::ustring& name, const Glib::ustring& value, bool) {
    std::ignore = name;
    bool result = Config::parse_input_data(value, this->memlimit);
    this->set_memlimit(this->memlimit); // Convert mebibytes to bytes
    return result;
  };
  og.add_entry(oe_mem, mh);

  oc.set_main_group(og);
  oc.parse(argv); // Use this instead of oc.parse(argc, argv) when argv is
                  // obtained from ApplicationCommandLine::get_arguments()

  if (version) {
    this->signal_print_info.emit(this->get_version_string());
    this->signal_exit.emit();
    return EXIT_SUCCESS;
  }

  if (g_strv_length(argv) > 2) { // Exit if two positional argument are given.
    this->signal_print_info.emit("Too many non-optional arguments!");
    this->signal_exit.emit();
    return EXIT_FAILURE;
  } else if (g_strv_length(argv) == 2)
      this->collection = Config::ConfigParser::parse_file(argv[1]);

  // Start event loop:
  this->signal_start_frontend.emit();

  g_strfreev(argv);
  return EXIT_SUCCESS;
}

} // namespace sinaps

#endif // SINAPS_BACKEND_H
