// Copyright 2023 xwst <xwst@gmx.net>

/*
This file is part of sinaps.

sinaps is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

sinaps is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with sinaps. If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef SINAPS_BACKEND_H_
#define SINAPS_BACKEND_H_

#include <list>
#include <memory>

#include <giomm.h>
#include <glibmm.h>

#include "backend/collection.h"
#include "secure_memory/passphrase.h"
#include "secure_memory/vector.h"

namespace sinaps {

class Backend {

  using GCV_uptr = SecureMemory::Vector_uptr<gunichar>;

  private:
    GCV_uptr pin;
    std::unique_ptr<SecureMemory::Passphrase<gunichar>> p;

  protected:
    unsigned long long opslimit;
    size_t memlimit;
    std::shared_ptr<Config::Collection> collection;

    static Glib::ustring get_version_string();

    bool argon2_parameters_ready();
    void initialize_passphrase_if_ready();
    bool handle_special_command(const Glib::ustring&);
    void handle_context_input(std::vector<Glib::ustring>&);

  public:

    Backend();

    /*****************************************************/
    /*********************** Slots ***********************/
    /*****************************************************/
    void on_frontend_ready();
    void set_pin(GCV_uptr&);
    void set_opslimit(unsigned long long);
    void set_memlimit(size_t);
    void on_frontend_closed();
    void on_line_entered(const Glib::ustring&);
    void on_input_buffer_changed(const Glib::ustring&);
    int parse_commandline_arguments(
      const Glib::RefPtr<Gio::ApplicationCommandLine>&
    );


    /*****************************************************/
    /********************** Signals **********************/
    /*****************************************************/

    sigc::signal<
      void(Glib::ustring, unsigned long long, size_t)
    > signal_argon2_parameters_missing;

    sigc::signal<void(std::list<Glib::ustring>)> signal_new_input_completions;

    sigc::signal<void(GCV_uptr&)> signal_password_to_clipboard;
    sigc::signal<void(void)> signal_clear_clipboard;

    sigc::signal<void(GCV_uptr&)> signal_display_password;
    sigc::signal<void(void)> signal_clear_output;

    sigc::signal<void(Glib::ustring)> signal_display_info;
    sigc::signal<void(Glib::ustring)> signal_print_info;

    sigc::signal<void(void)> signal_exit;
    sigc::signal<void(void)> signal_start_frontend;
    sigc::signal<void(void)> signal_start_input_loop;
};

} // namespace sinaps

#endif // SINAPS_BACKEND_H_
