// Copyright 2023 xwst <xwst@gmx.net>

/*
This file is part of sinaps.

sinaps is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

sinaps is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with sinaps. If not, see <http://www.gnu.org/licenses/>.
*/

#include "collection.h"

#include <string.h>

#include <glibmm.h>

#include "util.h"

/**
 * @class sinaps::Config::Collection
 * @brief This class represents a root node in the hierarchy of Context objects
 *        and its instances (including its children) correspond to a single
 *        configuration file.
 *
 * As opposed to the base class \ref sinaps::Config::ContextNode "ContextNode"
 * it can be constructed without a reference to a parent node and since this
 * class does not know about a parent, the get-functions of the member variables
 * return static default values if the optional member variables are not set.
 *
 * In addition to the default values, the Collection object defines the path
 * delimiter of the hierarchy which is stored in the member variable \ref name.
 */


namespace sinaps::Config {


/**
 * @brief Creates a Collection object using a mandatory path delimiter and a set
 *        of optional parameters.
 *
 * The optional parameters correspond to the class member variables of
 * ContextNode.
 *
 * @param delimiter The path delimiter used in the hierarchy.
 * @param alphabet See ContextNode::alphabet
 * @param counter See ContextNode::counter
 * @param length See ContextNode::length
 * @param skip See ContextNode::skip
 */
Collection::Collection(
  gunichar delimiter,
  std::optional<Glib::ustring> alphabet,
  std::optional<uint64_t> counter,
  std::optional<uint16_t> length,
  std::optional<uint32_t> skip
  ) : ContextNode(ContextNode::ctor_access_token(nullptr),
                  {
                    "",
                    alphabet.value_or(default_alphabet),
                    counter.value_or(default_counter),
                    length.value_or(default_length),
                    skip.value_or(default_skip)
                  }),
      delimiter(delimiter) {}


Glib::ustring Collection::get_path() {
  return this->params.name;
}


Glib::ustring Collection::get_alphabet() {
  return this->params.alphabet.value_or(default_alphabet);
}


uint64_t Collection::get_counter() {
  return this->params.counter.value_or(default_counter);
}


uint16_t Collection::get_length() {
  return this->params.length.value_or(default_length);
}


uint32_t Collection::get_skip() {
  return this->params.skip.value_or(default_skip);
}


gunichar Collection::get_path_delimiter() {
  return this->delimiter;
}


Glib::ustring Collection::config_string() {
  return this->config_string(Glib::ustring(1, this->get_path_delimiter()), 0);
}


Glib::ustring Collection::config_string(Glib::ustring delimiter, uint level) {
  Glib::ustring result;
  if (this->has_children()) {
    for(auto& child : this->children)
      result += delimiter + child->config_string(delimiter, level);
  } else result = delimiter;
  return result;
}


/**
 * @brief Returns a list of all children that match the given path prefix.
 *
 * @param prefix Path prefix that the path of the children need to start with.
 * @param include_parameters Whether the context parameters should be appended
 *                           to the path of each found node.
 * @return std::list< std::shared_ptr< sinaps::Config::ContextNode > >
 */
std::list<Glib::ustring> Collection::find_children(
  Glib::ustring prefix,
  bool include_parameters
) {
  auto result = std::list<Glib::ustring>();
  this->ContextNode::find_children(result, prefix, include_parameters);
  return result;
}


/**
 * @brief Return the context parameters for a child specified by path. If no
 *        matching child exists, a default initialized ContextNodeParameters is
 *        returned.
 */
ContextNodeParameters Collection::get_child_params(const Glib::ustring& path) {
  auto child = this->get_child_recursively(path);
  return (child == nullptr) ? ContextNodeParameters() : child->get_parameters();
}


}
