// Copyright 2023 xwst <xwst@gmx.net>

/*
This file is part of sinaps.

sinaps is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

sinaps is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with sinaps. If not, see <http://www.gnu.org/licenses/>.
*/

#include "contextnode.h"

#include <optional>

#include <glibmm.h>

/**
 * @brief The namespace for all classes related to the parsing of configuration
 *        files.
 *
 * ContextNode represent a context within a hierarchical tree-like structure.
 * All nodes are organized in a hierarchy where each node holds a pointer to its
 * parent and each parent stores its children in a list of shared pointers.
 *
 * A Collection is a special subclass that represents a root node and can be
 * constructed with nullptr as a parent.
 */
namespace sinaps::Config{

/**
 * @class ContextNode
 * @brief A hierarchy agnostic class storing context information for the
 *        derivation of passwords.
 *
 * This class contains fields required to derive a password for a particular use
 * case. Most parameters are optional to allow falling back to default values.
 * The private fields \ref parent and \ref children are used to establish a
 * tree-like hierarchy in which parents own their children.
 *
 * ContextNode objects can not be directly constructed but only created within
 * the hierarchy using add_child().
 *
 * For a more detailed description of the class member variables used in the
 * password derivation process see SecureMemory::Passphrase::custom.
 */

/**
  * @var ContextNodeParams::name
  * @brief The name of the context without the corresponding hierarchy.
  *
  * ContextNodes are organized in a tree like structure like a file systems and
  * this name represents the last part of the entire path (corresponding to a
  * file name).
  *
  * The full path can be obtained using \ref get_path().
  */

/**
 * @var ContextNode::parent
 * @brief A pointer to the parent node.
 *
 * If a particula member variable is not set, the parent is queried for its
 * value to allow value propagation down the hierarchy.
 */

/**
 * @var ContextNode::children
 * @brief A list storing shared pointers to the child nodes.
 *
 * The shared pointer is used to guarantee the existence of the children as long
 * as the parent exists. When a node is destructed, all its children may be
 * destructed, too.
 */

/**
  * @var ContextNodeParams::alphabet
  * @brief The characters used to encode the derived password.
  *
  * The derived password will be expressed in terms of all characters in the
  * parameter alphabet. The order and the multiplicity of the characters is
  * irrelevant.
  *
  * \sa SecureMemory::Passphrase::custom
  */

/**
  * @var ContextNodeParams::counter
  * @brief A counter which can be increased to generate a new password.
  *
  * \sa SecureMemory::Passphrase::custom
  */

/**
  * @var ContextNodeParams::length
  * @brief The length of the derived password (number of characters).
  *
  * \sa SecureMemory::Passphrase::custom
  *
  */

/**
  * @var ContextNodeParams::skip
  * @brief This parameter determines the number of characters to skip.
  *
  * The password derivation process generates a stream of characters from
  * which by default the first \ref length characters are drawn. If this
  * parameter is specified, the corresponding number of characters in the
  * stream will be omitted. The password length remains unchanged.
  *
  * \sa SecureMemory::Passphrase::custom
  */

/**
  * @fn ContextNode::get_name
  * @brief Getter for the name stored in the \ref params variable.
  *
  * @return Glib::ustring
  */

/**
 * @struct ContextNode::ctor_access_token
 * @brief An auxiliary struct used for access control of the ContextNode
 *        constructor.
 *
 * The ContextNode constructor is public to allow using std::make_shared but in
 * order to prevent other classes from creating ContextNode objects outside of a
 * hierarchy, the constructor requires an argument which is only accessible from
 * withing ContextNode.
 *
 * The constructor of ctor_access_token takes a pointer to a ContextNode as an
 * argument which is then passed to the parent argument of the ContextNode
 * constructor.
 */




/**
  * @brief Constructor with an argument for each class member variable.
  *
  * The pointer to the parent object is passed via the auxiliary
  * ctor_access_token.
  *
  * @param c An access token that prevents constructor usage from outside.
  * @param name ContextNode::name
  * @param alphabet ContextNode::alphabet
  * @param counter ContextNode::counter
  * @param length ContextNode::length
  * @param skip ContextNode::skip
  */
ContextNode::ContextNode(
  ContextNode::ctor_access_token c,
  const ContextNodeParameters& params
) : parent(c.parent), params(params) {}


/**
 * @brief Return the parameters of this node and replace unset values with the
 *        ones of the parent node.
 * @return sinaps::Config::ContextNodeParameters
 */
ContextNodeParameters ContextNode::get_parameters() {
  return {
    this->get_name(),
    this->get_alphabet(),
    this->get_counter(),
    this->get_length(),
    this->get_skip()
  };
}



/**
  * @brief Return a pointer to the child node with the name passed as an
  *        argument.
  *
  * If no child with the passed name exists, a nullptr is returned.
  *
  * @param child_name The name of the child which is queried.
  * @return ContextNode*
  */
ContextNode* ContextNode::get_child(const Glib::ustring& child_name) {
  for (auto iter = this->children.begin(); iter!=this->children.end(); ++iter) {
    if ((*iter)->get_name() == child_name) return iter->get();
  }
  return nullptr;
}


/**
 * @brief Finds and returns a child recursively. If the given path does not
 *        match an existing child, a nullptr is returned.
 */
ContextNode* ContextNode::get_child_recursively(const Glib::ustring& path) {
  auto delimiter = Glib::ustring(1, this->get_path_delimiter());
  auto path_pieces = Glib::Regex::split_simple(delimiter, path);
  ContextNode* child = this;
  for (auto piece : path_pieces) {
    child = child->get_child(piece);
    if (child == nullptr) break;
  }
  return child;
}


/**
  * @brief Returns true if the number of children is larger than zero.
  *
  * @return bool
  */
bool ContextNode::has_children() {
  return not this->children.empty();
}

/**
  * @brief Returns true if a child with the given name exists.
  *
  * @param child_name The name of the child in question.
  * @return bool
  */
bool ContextNode::contains(const Glib::ustring& child_name) {
  return this->get_child(child_name) != nullptr;
}

/**
  * @brief Creates a new ContextNode object and adds it as a child.
  *
  * If a child with the name already exists an error is thrown. If the function
  * is successful, a reference to the newly created child is returned.
  *
  * @return sinaps::Config::ContextNode&
  */
std::shared_ptr<ContextNode> ContextNode::add_child(
  const ContextNodeParameters& params
) {
  if (this->contains(params.name)) {
    auto err = "A child named '%1%2' already exists!";
    auto msg = Glib::ustring::compose(err, this->get_path(), params.name);
    throw std::runtime_error(msg);
  }

  auto ctor_token = ContextNode::ctor_access_token(this);
  auto child = std::make_shared<ContextNode>(ctor_token, params);
  this->children.push_back(child);
  return child;
}

/**
  * @brief Returns the alphabet used for encoding the password.
  *
  * @return Glib::ustring
  */
Glib::ustring ContextNode::get_alphabet() {
  return this->params.alphabet.value_or(this->parent->get_alphabet());
}

/**
  * @brief Returns the counter used in the password derivation.
  *
  * @return uint64_t
  */
uint64_t ContextNode::get_counter() {
  return this->params.counter.value_or(this->parent->get_counter());
}

/**
  * @brief Returns the length of the password.
  *
  * @return uint16_t
  */
uint16_t ContextNode::get_length() {
  return this->params.length.value_or(this->parent->get_length());
}

/**
  * @brief Returns the \ref skip parameter of the Context object.
  *
  * @return uint32_t
  */
uint32_t ContextNode::get_skip() {
  return this->params.skip.value_or(this->parent->get_skip());
}

/**
  * @brief Returns the path delimiter.
  */
gunichar ContextNode::get_path_delimiter() {
  return this->parent->get_path_delimiter();
}


/**
  * @brief Returns the path of the Context object including the names of all
  *        parents separated by a delimiter.
  *
  * If the Context obect has children itself, the a trailing delimiter is added
  * to the path.
  *
  * @return Glib::ustring
  */
Glib::ustring ContextNode::get_path() {
  auto base = this->parent->get_path();
  auto result = base + this->get_name();
  if (this->has_children()) result += this->get_path_delimiter();
  return result;
}


/**
 * @brief Auxiliary function to format a single optional argument.
 */
static void format_option(
  Glib::ustring& result,
  bool& key_required,
  bool has_value,
  Glib::ustring value,
  Glib::ustring name)
{
  if (has_value) {
    result += " ";
    if (key_required) result += name +"=";
    result += value;
    key_required = false;
  } else key_required = true;
}


/**
 * @brief Convert the optional parameters to a string that can be appended to
 *        the name to form a valid line in a configuration file.
 *
 * The result contains a leading blank.
 * @return Glib::ustring
 */
Glib::ustring ContextNode::optionals_to_string() {

  bool key = false;
  Glib::ustring result = "";

  format_option(result, key, this->params.length.has_value(),
                std::to_string(this->params.length.value_or(0)), "length");
  format_option(result, key, this->params.counter.has_value(),
                std::to_string(this->params.counter.value_or(0)), "counter");
  format_option(result, key, this->params.alphabet.has_value(),
                this->params.alphabet.value_or(""), "alphabet");
  format_option(result, key, this->params.skip.has_value(),
                std::to_string(this->params.skip.value_or(0)), "skip");

  return result;
}

/**
 * @brief Finds all children that match the given path prefix.
 *
 * All found matches are appended to the passed list of nodes.
 *
 * @param node_list The list of nodes to which matches are appended.
 * @param prefix Path prefix that the path of the children need to start with.
 * @param include_parameters Whether the context parameters should be appended
 *                           to the path of each found node.
 */
void ContextNode::find_children(
  std::list<Glib::ustring>& node_list,
  Glib::ustring prefix,
  bool include_parameters
) {
  for (auto c : this->children) {
    auto name = c->get_name();
    auto psize = prefix.size();
    if ((name.size() >= psize) and (prefix == name.substr(0, psize))) {
      if (c->has_children()) node_list.push_back(c->get_path());
      else node_list.push_back(
        ((include_parameters) ? c->to_string() : c->get_path()) + " "
      );
    } else if ((name.size() < psize)
               and (prefix.substr(0, name.size()) == name)
               and (prefix.at(name.size()) == this->get_path_delimiter()))
      c->find_children(
        node_list,
        prefix.substr(name.size() + 1),
        include_parameters
      );
  }
}

/**
 * @brief Convert the given node and all its children recursively to a string
 *        that can be parsed by ConfigParser.
 *
 * A single line of the result is not independently meaningful as optinal
 * arguments may need to be inherited by parent nodes.
 * A self-sufficient representation of a node can be obtained using to_string().
 *
 * @return Glib::ustring
 */
Glib::ustring ContextNode::config_string(Glib::ustring delimiter, uint level) {

  auto result = Glib::ustring(2*level, ' ');
  result += this->get_name();
  if (this->has_children()) {
    result += delimiter + this->optionals_to_string() + '\n';
    for(auto& child : this->children)
      result += child->config_string(delimiter, level + 1);
  } else result += this->optionals_to_string() + '\n';
  return result;
}


Glib::ustring ContextNode::to_string(bool use_keywords) {
  Glib::ustring r = this->get_path();
  r += (use_keywords ? " length=" : " ") + std::to_string(this->get_length());
  r += (use_keywords ? " counter=" : " ") + std::to_string(this->get_counter());
  r += (use_keywords ? " alphabet=" : " ") + this->get_alphabet();
  r += (use_keywords ? " skip=" : " ") + std::to_string(this->get_skip());
  return r;
}

}
