// Copyright 2023 xwst <xwst@gmx.net>

/*
This file is part of sinaps.

sinaps is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

sinaps is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with sinaps. If not, see <http://www.gnu.org/licenses/>.
*/

#include "parser.h"

#include <fstream>
#include <map>
#include <memory>

#include <glibmm.h>

#include "collection.h"

namespace sinaps::Config {


static auto nullopt = std::nullopt;


/**
  * @brief The default delimiter in the paths of context nodes.
  */
gunichar ConfigParser::delimiter = '/';


/**
 * @brief The symbol used to start comments in a configuration file.
 */
gunichar ConfigParser::comment = '#';


/**
 * @fn parse_input_data
 * @brief Convert a string to the datatype specified by the template parameter
 *        and store the correponding value in the second argument.
 *
 * @return true if the conversion succeeded and no input data is left in the
 *         string, false otherwise
 */


/**
 * @brief A special map<int, T> like container in which entries with larger key
 *        values are considered children of entries with smaller key values.
 * @tparam T The mapped data type
 *
 * As a consequence, when removing or replacing an entry with key K, all entries
 * with keys larger than K are automatically removed. This can be used to keep
 * track of the current position when traversing tree-like structures.
 */
template <typename T>
class ConfigParser::StackMap {

  private:
    std::map<int, T> data;

  public:
    /**
     * @brief Adds a key-value pair (or replaces an existing one) in the map and
     *        removes all subsequent content that is all keys that are larger
     *        than the one inserted or replaced.
     *
     * @param key The key to be inserted or replaced.
     * @param value The corresponding value.
     */
    void add_or_replace(int key, T value) {
      auto [iter, success] = data.insert_or_assign(key, value);
      data.erase(std::next(iter), data.end());
    };

    /**
     * @brief Return the value corresponding the largest key that is smaller
     *        than the argument.
     */
    T get_parent(int key) {
      return std::prev(data.lower_bound(key))->second;
    };
};


/**
 * @brief Remove trailing whitespaces from a string and returns the number of
 *        whitespaces removed.
 *
 * @param[in, out] s The string to be trimmed
 * @return uint
 */
uint ConfigParser::ltrim(Glib::ustring& s) {
  uint whitespace_count = 0;
  auto it = s.begin();
  while (it != s.end()) {
    if (std::isblank(*it)) {
      whitespace_count++;
      it++;
    } else break;
  }
  s = Glib::ustring(it, s.end());
  return whitespace_count;
}


template<typename T>
/**
 * @brief Parses an optional parameter in a line of a configuration file.
 *
 *
 * The template parameter T should be a numeric type that can be converted using
 * the std::stoull function.
 *
 * @param T The template referring to a numeric type.
 * @param in The string to be converted.
 * @param min The minimum allowed value. If the provided value is smaller than
 *            this, an exception is thrown.
 * @param field_name The name of the optional field. This variable is used for
 *                   error reporting, only.
 * @return T The numerical value of the provided string.
 */
T ConfigParser::str2int(const Glib::ustring& in, T min,
                             const std::string& field_name) {
    bool failed = false;
    T result;
    try {
        result = static_cast<T>(std::stoull(in.raw()));
    } catch(const std::logic_error& e) {
        failed = true;
    }
    if (in != Glib::ustring::sprintf("%ju", result)
        or result < min or failed) {
        auto err = "Error: Value '" + Glib::locale_from_utf8(in)
                  + "' of '" + field_name + "'-parameter"
                  + " is not in the allowed range.";
        throw std::domain_error(err);
    }
    return result;
}


// static void add_child_recursively(
//   Glib::ustring name,
//   std::optional<Glib::ustring> alphabet,
//   std::optional<uint64_t> counter,
//   std::optional<uint16_t> length,
//   std::optional<uint32_t> skip
// ) {
//   return;
// }


ConfigParser::ConfigParser() {
  this->levels = std::make_unique<StackMap<std::shared_ptr<ContextNode>>>();
  this->result = std::make_shared<Collection>(delimiter, nullopt, nullopt,
                                              nullopt, nullopt);
  this->levels->add_or_replace(-1, this->result);
}


/**
 * @brief Remove the comment symbol and all following characters from a string.
 *
 * @param[in,out] line The string in which comments are removed.
 */
void ConfigParser::remove_comments(Glib::ustring& line) {
  auto index = line.find(this->comment);
  if (index < Glib::ustring::npos)
    line.erase(index, line.size());
}


/**
 * @brief Convert an input line to a ContextNodeParameters-struct.
 *
 * The parameter can be obtained using Glib::Regex::split_simple("\\s+", line).
 *
 * @param ctx std::vector<Glib::ustring> of the fields in the entered line
 * @return sinaps::Config::ContextNodeParameters
 */
ContextNodeParameters ConfigParser::parse_context(
  std::vector<Glib::ustring>& ctx
) {

  ContextNodeParameters result;

  std::erase(ctx, "");
  result.name = ctx[0];

  std::array<Glib::ustring, 4> keys = {"length", "counter", "alphabet", "skip"};
  uint expected_key = 0;
  for (auto it = ctx.begin() + 1; it != ctx.end(); ++it) {
    auto index_eq = it->find_first_of('=');
    if (index_eq != Glib::ustring::npos) {
      auto key = it->substr(0, index_eq);
      *it = it->substr(index_eq+1, Glib::ustring::npos);
      expected_key = 0;
      while ((expected_key < keys.size()) and (keys[expected_key] != key))
        expected_key++;
    }
    if (expected_key >= keys.size())
      throw std::runtime_error(
        "Parsing error: Unnamed or unknown parameter after 'skip'-argument."
      );

    if (keys[expected_key] == "length")
      result.length = ConfigParser::str2int(*it, 1, "length");
    else if (keys[expected_key] == "counter")
      result.counter = ConfigParser::str2int(*it, 0, "counter");
    else if (keys[expected_key] == "alphabet")
      result.alphabet = *it;
    else if (keys[expected_key] == "skip")
      result.skip = ConfigParser::str2int(*it, 0, "skip");

    expected_key++;
  }

  return result;
}


void ConfigParser::parse_line(Glib::ustring& line) {
  this->remove_comments(line);
  auto indent = this->ltrim(line);
  if (line == "") return;

  auto line_content = Glib::Regex::split_simple("\\s+", line);

  auto parent = this->levels->get_parent(indent);
  auto args = this->parse_context(line_content);

  if (args.name.find(this->result->get_path_delimiter()) < Glib::ustring::npos)
    throw std::runtime_error(Glib::ustring::compose(
      "Parsing error: Names must not contain the path delimiter '%1'."
      "Offensive line:\n%2", (char) this->result->get_path_delimiter(), line
    ));

  auto child = parent->add_child(args);
  this->levels->add_or_replace(indent, child);
}


std::shared_ptr<Collection> ConfigParser::parse(std::string file_name) {
  std::ifstream file(file_name);
  std::string line;

  while (std::getline(file, line)) {
    auto ustring_line = Glib::ustring(line); // To do: Convert to unicode?
    this->parse_line(ustring_line);
  }

  return this->result;
}


std::shared_ptr<Collection> ConfigParser::parse_file(std::string file_name) {
  ConfigParser p = ConfigParser();
  return p.parse(file_name);
}

}
