// Copyright 2023 xwst <xwst@gmx.net>

/*
This file is part of sinaps.

sinaps is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

sinaps is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with sinaps. If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef SINAPS_FRONTEND_INTERFACE_H_
#define SINAPS_FRONTEND_INTERFACE_H_

#include <iostream>
#include <list>

#include <glibmm.h>
#include <sigc++/sigc++.h>

#include <secure_memory/vector.h>

namespace sinaps {

/**
  * @brief An interface for frontends that provides a set of signals and slots
  *        to be bound to the backend object.
  *
  * The frontend handles in- and output but does not derive passwords nor store
  * configuration file information. As a consequence, the backend needs to
  * provide a list of possible completions for the current input buffer if the
  * frontend should be able to support some sort of auto completion.
  * To that end, the frontend should emit the input_buffer_changed signal
  * whenever the user input changed and accept an update of the possible
  * completions via set_input_completions().
  */
class FrontendInterface {

  protected:

    using GCV_uptr = SecureMemory::Vector_uptr<gunichar>;

    virtual void close(const Glib::ustring& message) {
      this->display_info(message);
      this->signal_closed.emit();
      this->quit();
    };


  public:
    /*****************************************************/
    /*********************** Slots ***********************/
    /*****************************************************/
    /**
     * @brief Ask the user for Argon2 parameters.
     *
     * The function arguments specify which parameters need to be asked for as
     * described in the documentation of
     * Backend::signal_argon2_parameters_missing. Parameters that have already
     * been set can be displayed to the user.
     *
     * When the user aborts the input before all required parameters have been
     * successfully read, the interface should shut down. For each read
     * parameter the corresponding signal should be emitted.
     *
     * The prompt should avoid displaying the pin (by default) and guard the
     * input against swapping, etc..
     *
     * @sa Backend::signal_argon2_parameters_missing,@n
     *     @ref SecureMemory::Passphrase::Passphrase(
     *            Vector_uptr<T>,
     *            unsigned long long,
     *            size_t
     *          ) "Passphrase constructor",@n
     *     signal_pin_entered,@n
     *     signal_opslimit_entered,@n
     *     signal_memlimit_entered
     *
     * @param pin_info An informative string to be displayed when a pin is
     *                 already provided. If the string is empty, a pin has to be
     *                 entered.
     * @param current_opslimit The current value of opslimit.
     * @param current_memlimit The current value of memlimit.
     */
    virtual void prompt_argon2_parameters(
      const Glib::ustring& pin_info,
      unsigned long long current_opslimit,
      size_t current_memlimit
    ) = 0;


    /**
     * @brief Store the list of completions corresponding to the current input
     *        buffer.
     *
     * @param  std::list<Glib::ustring> The list of completions to be saved.
     */
    virtual void set_input_completions(std::list<Glib::ustring>) = 0;


    /**
     * @brief Display the password that is passed as an argument.
     *
     * The password is passed as a reference to a unique pointer as libsigc++
     * does not support move semantics. The frontend should invalidate the
     * unique pointer by taking ownership.
     *
     * @param SecureMemory::Vector_uptr<gunichar> The password in terms of a
     *        gunichar vector.
     * @sa Backend::signal_display_password
     */
    virtual void display_password(GCV_uptr&) = 0;


    /**
     * @brief Copy the password to the operating systems clipboard.
     *
     * The password is passed as a reference to a unique pointer as libsigc++
     * does not support move semantics. The frontend should invalidate the
     * unique pointer by taking ownership.
     *
     * @param SecureMemory::Vector_uptr<gunichar> The password in terms of a
     *        gunichar vector.
     * @sa Backend::signal_password_to_clipboard
     */
    virtual void password_to_clipboard(GCV_uptr&) = 0;


    /**
     * @brief Clear the content of the clipboard.
     *
     * This function should be called some time after password_to_clipboard().
     */
    virtual void clear_clipboard() = 0;


    /**
     * @brief Clear the output that was generated by display_password().
     */
    virtual void clear_output() = 0;


    /**
     * @brief Shutdown the frontend.
     */
    virtual void quit() = 0;


    /**
     * @brief Start continuously reading user input.
     *
     * The backend will call this function as soon as it has been properly set
     * up.
     */
    virtual void accept_user_input() = 0;


    /**
     * @brief Perform all steps required to get to a state in which the backend
     *        can send messages to are request input from the user.
     */
    virtual void start() = 0;


    /**
     * @brief Display information on the screen.
     *
     * Unlike print_info(), this function should make use of the user interface
     * to display the provided information.
     */
    virtual void display_info(const Glib::ustring& text) = 0;


    /**
     * @brief Print information on the command line.
     *
     * This function can be called to print version information or a short usage
     * description.
     * Unlike display_info(), this function always prints to the standard output
     * and works before the CLI or GUI is started.
     */
    virtual void print_info(const Glib::ustring& text) {
      std::cout << text << std::endl;
    }


    /*****************************************************/
    /********************** Signals **********************/
    /*****************************************************/
    /**
     * @brief Emit this signal as soon as the frontend is ready to prompt the
     *        user for input but before entering the main loop reading input
     *        data.
     */
    sigc::signal<void(void)> signal_ready;

    /**
     * @brief This signal should be omitted when the frontend is closed.
     *
     * A command line interface may support closing only via command (which is
     * handled by the backend), in which case this signal should not be emitted
     * at all.
     */
    sigc::signal<void(void)> signal_closed;


    /**
     * @brief A signal emitted whenever the user entered a line (by pressing
     *        enter, etc.).
     */
    sigc::signal<void(Glib::ustring)> signal_line_entered;


    /**
     * @brief This signal is required for automatic completion and needs to be
     *        emitted whenever the input line buffer changed (prior to
     *        submitting the input by pressing enter, etc.).
     */
    sigc::signal<void(Glib::ustring)> signal_input_buffer_changed;


    /**
     * @brief Emit this signal, when the user entered a pin.
     * @sa prompt_argon2_parameters
     */
    sigc::signal<void(GCV_uptr&)> signal_pin_entered;


    /**
     * @brief Emit this signal, when the users entered the amount of
     *        computations to perform in Argon2.
     * @sa prompt_argon2_parameters
     */
    sigc::signal<void(unsigned long long)> signal_opslimit_entered;


    /**
     * @brief Emit this signal, when the users entered the amount of
     *        memory to be used in Argon2.
     * @sa prompt_argon2_parameters
     */
    sigc::signal<void(size_t)> signal_memlimit_entered;
};

}

#endif // SINAPS_FRONTEND_INTERFACE_H_
