// Copyright 2023 xwst <xwst@gmx.net>

/*
This file is part of sinaps.

sinaps is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

sinaps is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with sinaps. If not, see <http://www.gnu.org/licenses/>.
*/

#include "frontend/readline_application.h"

#include <termios.h>
#include <unistd.h>

#include <iostream>

#include <giomm.h>
#include <glibmm.h>

#include "backend/parser.h"
#include "frontend/frontend_interface.h"
#include "secure_memory/vector.h"

namespace {


void set_echo_mode(bool on = true) {
    struct termios tty;
    tcgetattr(STDIN_FILENO, &tty);
    if(not on) tty.c_lflag = tty.c_lflag & ~ECHO;
    else tty.c_lflag = tty.c_lflag | ECHO;
    tcsetattr(STDIN_FILENO, TCSANOW, &tty);
}


}


namespace sinaps {


std::shared_ptr<char> ReadlineApplication::get_line(
  const char* prompt,
  bool save_history
) {
  return this->rp.get_line(prompt, save_history);
}


FrontendInterface::GCV_uptr ReadlineApplication::get_pin(int max_length) {
  sinaps::SecureMemory::Vector<gchar> pin; // Input buffer
  pin.resize(max_length + 1); // +1 for null termination

  // Read user input:
  std::cout << "Please enter pin: ";
  set_echo_mode(false);
  gsize input_len = std::cin.getline(pin.data(), pin.size()).gcount() - 1;
  set_echo_mode(true);
  std::cout << std::endl;

  // Get local character encoding
  std::string local_encoding;
  Glib::get_charset(local_encoding);

  // Create object for encoding conversion:
  Glib::IConv IC("UCS-4", local_encoding);

  // Output buffer
  auto result = std::make_unique<SecureMemory::Vector<gunichar>>();
  if(input_len == 0 or input_len == static_cast<gsize>(-1)) {
    this->close("No pin entered. Aborting.");
    return result;
  }
  result->resize(input_len); // Reserve one UCS4 character per input byte

  // Copy data pointers to pass indirect pointers to iconv():
  char* input  = reinterpret_cast<char*>(pin.data());
  char* output = reinterpret_cast<char*>(result->data());
  gsize output_size = result->size() * sizeof(gunichar);

  // Convert input to UCS4 encoding:
  size_t status = IC.iconv(&input, &input_len, &output, &output_size);
  if(status == static_cast<std::size_t>(-1)) {
    this->close("Conversion to UCS-4 encoding failed");
    return result;
  }

  // After executing iconv(), 'output_size' holds the number of remaining
  // bytes in 'result' that weren't written. Shrink vector to actual size:
  result->resize(result->size() - output_size / sizeof(gunichar));

  return result;
}



/**
 * @brief Read input data from user and convert it to a numeric value.
 *
 * @param[out] result A reference to a variable that will contain the result
 *                    after execution
 * @param[in] prompt A character string to show to the user when prompting
 * @param[in] minimum_value Minumum acceptable value
 * @return true if the input could be parsed and is larger than minimum_value,
 *         false otherwise
 */
template <typename T>
bool ReadlineApplication::get_numeric_value(
  T& result,
  const char* prompt,
  T minimum_value
) {
  std::shared_ptr<char> line = this->get_line(prompt, false);
  if (not line.get()) return false;
  bool status = Config::parse_input_data(line.get(), result);
  return status ? result >= minimum_value : false;
}


ReadlineApplication::ReadlineApplication()
  : Gio::Application(
      "sinaps.readline.Application",
      Gio::Application::Flags::HANDLES_COMMAND_LINE |
      Gio::Application::Flags::NON_UNIQUE
    ),
    rp(ReadlineProxy::get_instance()) {}


void ReadlineApplication::display_password(GCV_uptr& password) {
  Glib::ustring pw_string = Glib::ustring(password->begin(), password->end());
  password.reset();
  this->print_info(pw_string);
}


void ReadlineApplication::display_info(const Glib::ustring& text) {
  this->print_info(text);
}


void ReadlineApplication::start() {
  this->activate();
}


void ReadlineApplication::quit() {
  this->keep_running = false;
  this->release();
}


void ReadlineApplication::clear_output() {
    printf("\033[2J\033[1;1H");
}


void ReadlineApplication::clear_clipboard() {
  // todo
}


void ReadlineApplication::password_to_clipboard(GCV_uptr& password) {
  Glib::ustring pw_string = Glib::ustring(password->begin(), password->end());
  password.reset();
  // todo
}


void ReadlineApplication::set_input_completions(
  std::list<Glib::ustring> completions
) {
  this->rp.set_completions(completions);
}


void ReadlineApplication::on_activate() {
  this->signal_ready.emit();
}


void ReadlineApplication::accept_user_input() {
  this->rp.input_buffer_changed.connect(
    this->signal_input_buffer_changed.make_slot()
  );

  printf("\u001B[?1049h\u001B[H"); // switch to alternate screen
  std::shared_ptr<char> line;
  this->keep_running = true;

  while(this->keep_running) {
    line = this->rp.get_line("sinaps> ", true);
    if(not line.get()) { // CTRL + d
      this->close("");
      continue;
    }
    if(not *line) continue; // empty line

    this->signal_line_entered.emit(Glib::ustring(line.get()));
  }

  printf("\u001B[?1049l"); // switch back from alternate screen
}


void ReadlineApplication::prompt_argon2_parameters(
  const Glib::ustring& pin_info,
  unsigned long long current_opslimit,
  size_t current_memlimit
) {
  if (current_opslimit == 0) {
    unsigned long long opslimit;
    auto prompt = "Enter the number of iterations: ";
    if (not this->get_numeric_value(opslimit, prompt, current_opslimit + 1)) {
      this->close("Could not parse input!");
      return;
    }
    else this->signal_opslimit_entered.emit(opslimit);
  }
  if (current_memlimit == 0) {
    size_t memlimit;
    auto prompt = "Amount of memory to be used [MiB]: ";
    if (not this->get_numeric_value(memlimit, prompt, current_memlimit + 1)) {
      this->close("Could not parse input!");
      return;
    }
    else this->signal_memlimit_entered.emit(memlimit);
  }
  if (pin_info.size() == 0){
    auto result = get_pin(1024);
    if (result->size() > 0)
      this->signal_pin_entered.emit(result);
  }
}

}
