// Copyright 2023 xwst <xwst@gmx.net>

/*
This file is part of sinaps.

sinaps is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

sinaps is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with sinaps. If not, see <http://www.gnu.org/licenses/>.
*/

#include "readline_proxy.h"

#include <iterator>

#include <readline/history.h>
#include <readline/readline.h>


namespace {

/**
 * @brief Auxiliary function to replace the first occurrence of a character in a
 *        C-style string.
 */
void replace_first_character(char* str, char old_char, char new_char) {
    char* location = strchr(str, old_char);
    if (location != nullptr) *location = new_char;
}

/**
 * @brief Removes the last part of a path, which is either the file name or the
 *        directory name followed by the path delimiter.
 *
 * The passed string is modified inplace.
 */
void remove_directory_part(char* str, char delimiter) {
    int last_delimiter_pos = -1;
    for (int i=strlen(str) - 2; i>=0; i--) {
        if (str[i] == delimiter) {
            last_delimiter_pos = i;
            break;
        }
    }
    if (last_delimiter_pos < 0) return;
    for (int i=last_delimiter_pos; static_cast<size_t>(i)<strlen(str); i++)
        str[i - last_delimiter_pos] = str[i + 1];
        // "+1" on rhs. to skip path delimiter but include '\0'
}


/**
 * @brief Auxiliary function to override the behaviour of readline when
 *        displaying a list of possible completions.
 *
 * This function removes the first whitespace and all following characters from
 * the strings to be displayed before invoking readlines default function for
 * showing a list of possible completions.
 * Reasoning: The actual completion includes the optional parameters (length,
 * counter, ...) but these parameters should not be shown in the list of
 * matches.
 */
void readline_custom_completion_display(
    char **matches,
    int num_matches,
    int max_length
) {
  max_length = 0;
  for (int i=0; i<=num_matches; i++) {
    replace_first_character(matches[i], ' ', '\0');
    remove_directory_part(matches[i], '/');
    auto length = static_cast<int>(strlen(matches[i]));
    if (length > max_length) max_length = length;
  }
  rl_display_match_list(matches, num_matches, max_length);
  rl_on_new_line();
  rl_redisplay();
}

} // namespace

/**
 * @class sinaps::ReadlineProxy
 * @brief A proxy class to deal with the readline library.
 *
 * This class provides a single function to read a line of user input using the
 * readline-library and provides methods to set up tab-completion. In order to
 * make use of tab-completion a list of possible completions must be set using
 * set_completions(). The signal input_buffer_changed() is emitted when the user
 * input has changed and can be used to trigger an update of the list of
 * possible completions.
 *
 * Since the readline library uses global variables and does not know about
 * member functions in C++, this class is implemented using the singleton
 * pattern.
 */

/**
 * @var sinaps::ReadlineProxy::input_buffer_changed
 * @brief A signal that is emitted whenever the user input changed.
 */

namespace sinaps {

/**
 * @brief Private constructor that sets up the connection to readline by
 *        storing function pointers in the appropriate global variables.
 *
 * @sa get_instance()
 */
ReadlineProxy::ReadlineProxy() {
  rl_redisplay_function = ReadlineProxy::redisplay_hook;
  rl_attempted_completion_function = ReadlineProxy::tab_completion;
  rl_completion_display_matches_hook = readline_custom_completion_display;
  std::ignore = readline_custom_completion_display;
}


/**
 * @brief Tests if the input buffer has changed by comparing with a stored copy.
 *
 * When the buffer has changed, a copy of the new buffer is stored for future
 * calls.
 *
 * @return bool
 */
bool ReadlineProxy::has_input_buffer_changed() {
  auto new_buffer = Glib::ustring(rl_line_buffer);
  if (new_buffer != this->current_line_buffer) {
    this->current_line_buffer = new_buffer;
    return true;
  }
  return false;
}


/**
 * @brief This function is invoked by readline every time the display is updated
 *        with the current content of the line buffer.
 *
 * It calls the default redisplay-function of readline but emits the
 * input_buffer_changed signal if the input buffer has really changed.
 */
void ReadlineProxy::redisplay_hook() {
  auto& instance = ReadlineProxy::get_instance();
  if (instance.has_input_buffer_changed())
    instance.input_buffer_changed.emit(instance.current_line_buffer);
  rl_redisplay(); // readline's built-in default implementation
}


/**
 * @brief Static singleton accessor.
 *
 * The copy constructor is delete so the result of this function needs to be
 * assigned to a reference.
 *
 * @return std::shared_ptr< sinaps::ReadlineProxy >
 */
ReadlineProxy& ReadlineProxy::get_instance() {
  static ReadlineProxy singleton_instance;
  return singleton_instance;
}


/**
 * @brief Custom generator for readlines completion functionality.
 *
 * This function assumes that readline_completions is properly set up when
 * entering. This requires a proper handling of the input_buffer_changed signal
 * with an immediate update using set_completions.
 *
 * Due to the number of calls by the readline library, this funcion must not
 * change the list of possible completions (i.e. by using pop_front()) and picks
 * the match to return based on the value of state.
 *
 * @sa https://tiswww.case.edu/php/chet/readline/readline.html#Completion-Variables
 */
char* ReadlineProxy::completion_entry_func(const char* text, int state) {
  std::ignore = text;
  std::ignore = state;
  auto& instance = ReadlineProxy::get_instance();

  if (static_cast<uint>(state) < instance.readline_completions.size()) {
    auto it = instance.readline_completions.begin();
    std::advance(it, state);
    return strdup(it->c_str());
  } else return nullptr;
}


/**
 * @brief Custom completion function for readline.
 *
 * @sa https://tiswww.case.edu/php/chet/readline/readline.html#Completion-Variables
 */
char** ReadlineProxy::tab_completion(const char* text, int start, int end) {
  std::ignore = end; // unused parameter is predefined by readline library
  rl_attempted_completion_over = 1;
  rl_completion_suppress_append = 1;
  if (start == 0)
    return rl_completion_matches(text, ReadlineProxy::completion_entry_func);
  return nullptr;
}


/**
 * @brief Use the readline library to prompt for a user input.
 *
 * @param prompt The prompt string to print when reading input.
 * @param save_history Whether to save the input to this readline history.
 * @return std::shared_ptr< char >
 */
std::shared_ptr<char> ReadlineProxy::get_line(
  const char* prompt,
  bool save_history
) {
    auto line = std::shared_ptr<char>(readline(prompt));
    if(line.get() && *line.get() && save_history) add_history(line.get());
    return line;
}

/**
 * @brief Replace the list of possible completions for the current line buffer
 *        with the passed ones.
 */
void ReadlineProxy::set_completions(std::list<Glib::ustring> completions) {
  this->readline_completions = completions;
}

};
