// Copyright 2023 xwst <xwst@gmx.net>

/*
This file is part of sinaps.

sinaps is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

sinaps is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with sinaps. If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef SINAPS_SECURE_MEMORY_CONSTANTS_H_
#define SINAPS_SECURE_MEMORY_CONSTANTS_H_

#include <sodium.h>

/**
 * \file
 * \brief Collection of customizable constants that modify the password
 *        derivation process.
 *
 * This file only includes constants that can be changed without changing any
 * implementation details. Constants that require a change of the code base when
 * modified are defined in the units where they are used (for instance
 * util::PASSWORD_SEED_LENGTH).
 */

namespace sinaps::SecureMemory {

    /**
     * \brief Version of the applied Argon2-algorithm
     */
    #define ARGON_ALGORITHM crypto_pwhash_ALG_ARGON2ID13

    /**
     * \brief Length in bytes of passphrase computed by the Argon2-algorithm
     */
    static const size_t PASSPHRASE_LEN = 32; // 256 bit

    /**
     * \brief Default salt for libsodium crypto_pwhash() function
     *
     * Makes sure, that the derived passwords are application dependent. A
     * random per user salt would be better, but is in conflict with the
     * requirement, that the passwords are derived only from a single main
     * pin. To mitigate this, the user should be encouraged, to pick a
     * reasonable strong main pin.
     */
    static const char* const MAIN_SALT = "sinaps is not a password store";

} // namespace sinaps::SecureMemory

#endif // SINAPS_SECURE_MEMORY_CONSTANTS_H_
