// Copyright 2023 xwst <xwst@gmx.net>

/*
This file is part of sinaps.

sinaps is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

sinaps is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with sinaps. If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef SINAPS_SECURE_MEMORY_PASSPHRASE_H_
#define SINAPS_SECURE_MEMORY_PASSPHRASE_H_

#include <cmath>
#include <cstring>
#include <ostream>
#include <set>
#include <vector>

#include <sodium.h>

#include "secure_memory/constants.h"
#include "secure_memory/util.h"
#include "secure_memory/vector.h"

namespace sinaps::SecureMemory {

template <typename T>
class Passphrase {

  public:
    Passphrase(const Passphrase&) = delete;
    Passphrase& operator=(Passphrase&&) = delete;
    Passphrase(Passphrase&&) = delete;
    Passphrase() = delete;

    Passphrase(Vector_uptr<T>, unsigned long long, size_t);
    ~Passphrase();

    void erase();

    std::vector<unsigned char> hash();

    Vector_uptr<T> custom(
      const std::vector<T>&,
      uint64_t,
      uint32_t,
      const std::vector<T>&,
      uint32_t = 0
    ) const;
    void writeRaw(
      std::ostream&,
      const std::vector<T>&,
      uint64_t,
      uint32_t,
      uint32_t = 0
    ) const;

  private:
    unsigned char* data;
    static const size_t len = PASSPHRASE_LEN;

    void setNoAccess() const;
    void setReadOnly() const;

    UCV_uptr deriveSeed(const std::vector<T>&) const;

  };

} // namespace sinaps::SecureMemory


/**
 * \class sinaps::SecureMemory::Passphrase
 * \brief Represents a passphrase derived from a user provided pin using the
 *        hash function Argon2
 *
 * \paragraph motivation Motivation
 * The general problem with user passwords is that they are usually low in
 * entropy and often not suitable for authentication, let alone cryptography.
 * This can be worked around to some extend by applying a password hashing
 * function that is computationally expensive, so that brute force attacks are
 * less viable.
 *
 * The Argon2 function is a modern example for such password hashing functions.
 * It accepts a user pin and parameters that control the running time and memory
 * consumption and produces a long passphrase. The output length may be
 * specified in general; if it is suitably large, the output can not be brute
 * forced. While the entropy did not change (the search space of user pins is
 * the same), it is much harder to brute force the user pin, as it is required
 * to perform the Argon2 computation for every possible user pin.
 *
 * Instead of using the output of Argon2 directly as a password, it can be used
 * to derive multiple passwords. That way, the user has to remember only one pin
 * and the computationally expensive hashing needs to performed only once. If
 * the pin is sufficiently complex (depends on the chosen Argon2 parameters),
 * the hashed passphrase will be cryptographically secure and all derived
 * passwords will be, too.
 *
 * This class is a wrapper for the passphrase data (stored in
 * \link Passphrase::data data\endlink) obtained from the Argon2 hashing
 * function. It can be only constructed from a user provided pin and Argon2
 * parameters. Once constructed, the stored passphrase data can not be altered.
 * Additionally, it can not be read directly, but class methods allow to derive
 * passwords from the passphrase. Read access can be additionally restricted
 * using one of the methods setNoAccess() and setReadOnly(), which prevents
 * unintended and malicious access.
 *
 * \paragraph derivation Password Derivation
 * Derived passwords are obtained by hashing the Passphrase::data
 * with a context string and a counter. That way, a single passphrase (or rather
 * user pin) can be used to derive passwords for different purposes (contexts).
 * The counter can be increased, if it is necessary to generate a new password
 * for a certain purpose. Since the passphrase data is considered to be
 * cryptographically secure (the user has to choose a sufficiently strong pin),
 * the derivation of the password can be done using generic hash algorithms.
 *
 * The generic hash algorithm provided by the libsodium library is restricted to
 * a maximum length of 64 bytes (crypto_generichash_BYTES_MAX). For special use
 * cases, this might be insufficient. Hence, context is hashed using the
 * passphrase data as a key (keyed hash) to obtain a seed (see deriveSeed()).
 * The seed is fed into the stream cipher ChaCha20 to produce a password of
 * (almost) arbitrary lenght. The counter is used as a nonce for ChaCha20. The
 * output out ChaCha20 is binary data and needs to be encoded to be used as a
 * password in frontend applications.
 *
 * Seed derivation, password generation and encoding is combined in the
 * following functions:
 * - custom()
 * - writeRaw()
 */

/**
 * \var Passphrase::data
 * \brief Stores the passphrase data obtained from Argon2.
 *
 * The memory is allocated using \b sodium_malloc() to prevent swapping its
 * content and allow enforcing read restrictions. It needs to be freed with
 * \b sodium_free() which also erases its content.
 *
 * \sa Passphrase::setNoAccess() and Passphrase::setReadOnly()
 */

/**
  * \var Passphrase::len
  * \brief Length of the passphrase \link Passphrase::data data\endlink
  *        in bytes.
  *
  * Value is copied from ::PASSPHRASE_LEN.
  */

namespace sinaps::SecureMemory {

/**
  * \brief Constructs an object from a user \a pin and the Argon2 parameters
  *        \a opslimit and \a memlimit.
  *
  * Same parameters always lead to the same passphrase, but changing any of
  * the parameters (including \a opslimit or \a memlimit) yields a different
  * passphrase.
  *
  * \param [in] pin a pin that was entered by the user
  * \param [in] opslimit one of the Argon2 parameters. Increasing \a opslimit
  *             increases the time spent to compute the passphrase
  * \param [in] memlimit number of bytes of memory used by the Argon2 funcion
  *
  * \sa the [libsodium documentation]
  *     (https://doc.libsodium.org/password_hashing/default_phf) for details
  *     regarding the Argon 2 parameters
  */
template <typename T>
Passphrase<T>::Passphrase(
  Vector_uptr<T> pin,
  unsigned long long opslimit,
  size_t memlimit
) {

  if (
    opslimit < crypto_pwhash_OPSLIMIT_MIN ||
    opslimit > crypto_pwhash_OPSLIMIT_MAX ||
    memlimit < crypto_pwhash_MEMLIMIT_MIN ||
    memlimit > crypto_pwhash_MEMLIMIT_MAX
  )
    throw std::domain_error("Invalid parameters for 'crypto_pwhash': "
                            "opslimit=" + std::to_string(opslimit) +
                            ", memlimit=" + std::to_string(memlimit));

  // Use hash function to create fixed size salt from salt string:
  unsigned char salt_hash [crypto_pwhash_SALTBYTES];
  crypto_generichash(salt_hash, crypto_pwhash_SALTBYTES,
                     reinterpret_cast<const unsigned char*>(MAIN_SALT),
                     std::strlen(MAIN_SALT), NULL, 0);

  // Derive passphrase data:
  this->data = static_cast<unsigned char*>(sodium_malloc(Passphrase::len));
  int status
    = crypto_pwhash(
        this->data,
        Passphrase::len,
        reinterpret_cast<const char*>(pin->data()),
        sizeof(T)*pin->size(),
        salt_hash,
        opslimit,
        memlimit,
        ARGON_ALGORITHM
      );

  if(status != 0) {
    this->erase();
    throw std::runtime_error("Could not compute passphrase.");
  }
  this->setNoAccess();
}

/**
  * \brief Destructor: Frees (and erases) the passphrase
  *        \link Passphrase::data data\endlink.
  */
template <typename T>
Passphrase<T>::~Passphrase() {
  sodium_free(this->data);
}

/**
  * \brief Erases the passphrase \link Passphrase::data data\endlink
  *        but does not delete the object.
  *
  * Not particularly useful at the moment, as the data can not be overwritten
  * by meaningful data and will stay in the "erased" state.
  */
template <typename T>
void Passphrase<T>::erase() {
  sodium_memzero(this->data, Passphrase::len);
}



/**
  * \brief Prevents access to the passphrase
  *        \link Passphrase::data data\endlink entirely.
  *
  * This is useful, if the data is not supposed to be read: Unintended and
  * malicious access attempts will be cancelled by aborting the program.
  *
  * \sa setReadOnly()
  */
template <typename T>
void Passphrase<T>::setNoAccess() const {
  sodium_mprotect_noaccess(this->data);
}

/**
  * \brief Allows readonly access to the passphrase
  *        \link Passphrase::data data\endlink.
  *
  * \sa the counter part setNoAccess()
  */
template <typename T>
void Passphrase<T>::setReadOnly() const {
  sodium_mprotect_readonly(this->data);
}

/**
  * \brief Derives a password using \a context, \a counter and
  *        \link Passphrase::data data\endlink and encodes it using the
  *        characters provided in \a alphabet.
  *
  * The password is derived using deriveSeed() and
  * util::derivePasswordBlock(). It is encoded using the characters passed in
  * \a alphabet but not null-terminated. The provided characters are sorted
  * and duplicates are removed prior to encoding.
  *
  * The first \a skip bytes will be skipped, i.e. concatenating the results
  * of custom(length=10, skip=0) and custom(length=10, skip=10) yields the
  * same as invoking custom(length=20, skip=0). This can be used to compute a
  * long password piecewise.
  *
  * This function combines util::PW_BLOCK_SIZE bytes of pseudo random data to
  * a single character, while writeRaw() provides access to the underlying
  * stream of raw bytes.
  *
  * \param [in] context user provided string for password derivation. Could
  *                     be a website a user name, the WiFi SSID, etc.
  * \param [in] counter may be used to generate a new password for a given
  *                     context by increasing by one
  * \param [in] length  number of characters in the derived password
  * \param [in] alphabet the set of characters that will appear in the return
  *                      value. Duplicates are removed.
  * \param [in] skip    Omit the first \a skip characters (allows random
  *                     access). Note, that \a length + \a skip still needs
  *                     to be less than \a UINT32_MAX
  * \return a password without null-termination that contains \a length
  *         characters of the string \a alphabet
  */
template <typename T>
Vector_uptr<T> Passphrase<T>::custom(
  const std::vector<T>& context,
  uint64_t counter,
  uint32_t length,
  const std::vector<T>& alphabet,
  uint32_t skip
) const {

  if(length > UINT32_MAX - skip)
    throw std::domain_error("'length + skip' must be less or equal to"
                            " 'UINT32_MAX'.");

  // Create alphabet vector
  // Keep each occurring character only one time and sort characters:
  const std::set<T> characters(alphabet.begin(), alphabet.end());
  if(characters.size() < 2)
    throw std::domain_error("The given alphabet needs to contain at "
                            "least two distinct characters.");

  // Allows index-based access:
  const std::vector<T> alphabet_vector(characters.begin(),
                                       characters.end());
  unsigned long int N = alphabet_vector.size(); // Number of characters

  UCV_uptr seed = this->deriveSeed(context);

  auto result = std::make_unique<Vector<T>>();
  auto block = std::make_unique<UCV>();
  result->reserve(length);
  block->resize(util::PW_BLOCK_SIZE);

  std::vector<unsigned char> nonce = util::counter2nonce(counter);

  for(unsigned int i=skip; i<skip+length; ++i) {
    util::derivePasswordBlock(*seed, nonce, i, *block);
    unsigned int char_index = util::block_modulo(*block, N);
    result->push_back(alphabet_vector[char_index]);
  }

  return result;
}

/**
  * \brief Derives a raw binary password using \a context, \a counter and
  *        \link Passphrase::data data\endlink and writes the raw
  *        data to the given ostream.
  *
  * The password is derived using deriveSeed() and
  * util::derivePasswordBlock(). It is not encoded and the raw data will be
  * written to the provided ostream object.
  *
  * The first \a skip bytes will be skipped, i.e. concatenating the results
  * of custom(length=10, skip=0) and custom(length=10, skip=10) yields the
  * same as invoking custom(length=20, skip=0). This can be used to compute a
  * long password piecewise.
  *
  * This function provides access to a raw stream of pseudo random data,
  * while custom() combines util::PW_BLOCK_SIZE bytes to a single character.
  *
  * \param [in] out     the stream to which the derived bytes are written
  * \param [in] context user provided string for password derivation. Could
  *                     be a website a user name, the WiFi SSID, etc.
  * \param [in] counter may be used to generate a new password for a given
  *                     context by increasing by one
  * \param [in] length  number of bytes in the derived password
  * \param [in] skip    Omit the first \a skip bytes (allows random access).
  *                     Note, that \a length + \a skip still needs to be less
  *                     than \a UINT32_MAX
  */
template <typename T>
void Passphrase<T>::writeRaw(
  std::ostream& out,
  const std::vector<T>& context,
  uint64_t counter,
  uint32_t length,
  uint32_t skip
) const {

  if(length > UINT32_MAX - skip)
    throw std::domain_error("'length + skip' must be less or equal to"
                            "'UINT32_MAX'.");

  UCV_uptr seed = this->deriveSeed(context);

  auto block = std::make_unique<UCV>();
  block->resize(util::PW_BLOCK_SIZE);

  std::vector<unsigned char> nonce = util::counter2nonce(counter);

  unsigned int block_counter = skip / util::PW_BLOCK_SIZE;
  unsigned int first = skip % util::PW_BLOCK_SIZE;

  while(length > 0) {
    util::derivePasswordBlock(*seed, nonce, block_counter, *block);

    uint32_t bytesToWrite = std::min(static_cast<unsigned long>(length),
                                     util::PW_BLOCK_SIZE-first);
    out.write(reinterpret_cast<char*>(block->data() + first), bytesToWrite);

    length -= bytesToWrite;
    first = 0;
    block_counter += 1;
  }
}

/**
  * \brief Derives a seed from the passphrase
  *        \link Passphrase::data data\endlink and a context string,
  *        which is used for the password generation.
  *
  * The returned seed is util::PW_SEED_LENGTH bytes long, so that it can be
  * used in util::derivePassword() and util::derivePasswordBlock().
  *
  * \param [in] context user provided string that is hashed with
  *             \link Passphrase::data data\endlink
  * \return a cryptographically strong seed that can be used in CSPRNG to
  *         compute passwords
  */
template <typename T>
UCV_uptr Passphrase<T>::deriveSeed(const std::vector<T>& context) const {

  auto result = std::make_unique<UCV>();
  result->resize(util::PW_SEED_LENGTH);
  this->setReadOnly();
  int status
    = crypto_generichash(
        result->data(),
        util::PW_SEED_LENGTH,
        reinterpret_cast<const unsigned char*>(context.data()),
        sizeof(T)*context.size(),
        this->data,
        Passphrase::len
      );
  this->setNoAccess();
  if(status != 0) throw std::runtime_error("Failed to derive seed using "
                                            "'crypto_generichash'.");

  return result;
}

} // namespace sinaps::SecureMemory


#endif // SINAPS_SECURE_MEMORY_PASSPHRASE_H_
