// Copyright 2023 xwst <xwst@gmx.net>

/*
This file is part of sinaps.

sinaps is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

sinaps is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with sinaps. If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef SINAPS_SECURE_MEMORY_SODIUMALLOCATOR_H_
#define SINAPS_SECURE_MEMORY_SODIUMALLOCATOR_H_

#include <limits>
#include <stdexcept>
#include <tuple>

#include <sodium.h>

/**
 * \file
 * \brief Definition of custom allocator that uses libsodiums 'sodium_malloc'.
 *
 * Used with STL containers, so that the memory managed by the container is
 * guarded against beeing swapped and erased properly when freed.
 */

namespace sinaps::SecureMemory {

template <class T> class SodiumAllocator;

/**
 * \brief Void specialization of SodiumAllocator
 */
template <> class SodiumAllocator<void> {
    public:
        typedef void*       pointer;
        typedef const void* const_pointer;

        typedef void value_type;
        template <class U> struct rebind { typedef SodiumAllocator<U> other; };
};

/**
 * \brief Custom allocator to use sodium's protected memory allocation with STL containers.
 */
template <class T> class SodiumAllocator {
    public:
        typedef T                 value_type;
        typedef std::size_t       size_type;
        typedef std::ptrdiff_t    difference_type;
        typedef value_type*       pointer;
        typedef const value_type* const_pointer;
        typedef value_type&       reference;
        typedef const value_type& const_reference;


        template <class U> struct rebind { typedef SodiumAllocator<U> other; };


        inline explicit SodiumAllocator() {
            if(sodium_init() == -1) {
                throw std::runtime_error("Could not initialize libsodium.");
            }
        }
        inline explicit SodiumAllocator(const SodiumAllocator&) {};
        template <class U> SodiumAllocator(const SodiumAllocator<U>&) {}
        ~SodiumAllocator() {}


        inline pointer address(reference x) { return &x; }
        inline const_pointer address(const_reference x) { return &x; }


        inline pointer allocate(size_type n,
                                    typename SodiumAllocator<void>
                                             ::const_pointer hint = 0
                               ) {
            std::ignore = hint;
            return static_cast<pointer>(sodium_malloc(n * sizeof(value_type)));
        }
        inline void deallocate(pointer p, size_type) {
            sodium_free(p);
        }


        inline size_type max_size() {
            return std::numeric_limits<size_type>::max() / sizeof(value_type);
        }


        inline bool operator==(SodiumAllocator const&) { return true; }
        inline bool operator!=(SodiumAllocator const&) { return false; }
    };

} // namespace sinaps::SecureMemory

#endif // SINAPS_SECURE_MEMORY_SODIUMALLOCATOR_H_
