// Copyright 2023 xwst <xwst@gmx.net>

/*
This file is part of sinaps.

sinaps is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

sinaps is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with sinaps. If not, see <http://www.gnu.org/licenses/>.
*/

#include "secure_memory/util.h"

#include <algorithm>
#include <charconv>
#include <cmath>
#include <set>
#include <vector>


/**
 * \namespace sinaps::SecureMemory::util
 * \brief %Collection of utility functions used by Passphrase
 */

/**
 * \var util::PW_SEED_LENGTH
 * \brief Number of bytes that need to be passed to derivePasswordBlock() as
 *        seed for the password generation.
 *
 * Its value depends on the implementation and might change if the password
 * derivation prescription is changed.
 */

/**
 * \var util::PW_BLOCK_SIZE
 * \brief Size of password blocks that are computed by derivePasswordBlock().
 *
 * Currently, it is assigned the size of a ChaCha20-block, which is 64 byte.
 * Although it might be reasonable performance wise to choose an integer
 * multiple of the ChaCha20-blocksize, its value could be anyhing if the
 * implementation of derivePasswordBlock() is adjusted appropriately.
 */

namespace sinaps::SecureMemory::util {

    /**
     * \brief Converts a sequence of bytes to an ASCII character array of zeros
     *        and ones. The array is not null-terminated.
     *
     * Each byte is represented by eight digits, each of which corresponds to
     * one bit. The result is formatted in a human readable format, that is the
     * leading digit is stored in the first element of the data array.
     *
     * \param [in] passphrase data to be converted
     * \return the input converted to base 2 as an array of the characters '0'
     *         and '1'
     */
    UCV_uptr asciiBitString(const UCV& passphrase) {
        const int binary_size = passphrase.size() * 8;
        auto result = std::make_unique<UCV>();
        result->resize(binary_size);

        int masks[] = {1, 2, 4, 8, 16, 32, 64, 128};
        for(int i=0; i<binary_size; ++i) {
            int byte_index = passphrase.size() - 1 - i/8;
            if((passphrase.data()[byte_index] & masks[i % 8]) >= 1)
                 result->data()[binary_size - 1 - i] = '1';
            else result->data()[binary_size - 1 - i] = '0';
        }

        return result;
    }

    /**
     * \brief Converts an unsigned integer to a platform independent binary
     *        representation
     *
     * This function computes the binary representation of the given integer.
     * The size of the result is specified using the second parameter. If the
     * requested size is larger than the length of the binary number, the data
     * is padded with trailing zeros. If the requested size is to small, an
     * exeption is thrown. It is up to the caller, to guarantee, that
     * \a byte_len is sufficiently large.
     *
     * The return value is a std::vector of unsigned char without null
     * termination. It is independent of the binary representation of the given
     * integer (endianness).
     *
     * This function is used to convert an integer to a platform independent
     * character array of given length. This is used to convert a counter to a
     * nonce for password generation, which ensures, that increasing the counter
     * yields a new password, even if the pin and context remain unchanged.
     *
     * \sa counter2nonce()
     *
     * \param [in] N integer to be converted
     * \param [in] byte_len number of characters of the result
     * \return vector of unsigned characters that represent the given integer in
     *         binary without null termination. The size of the vector equals
     *         \a byte_len.
     */
    std::vector<unsigned char> binaryRepresentation(uint64_t N,
                                                    size_t byte_len) {

        // length of hex representation:
        size_t hex_len = (N == 0) ? 1 : static_cast<size_t>(log2(N) / 4) + 1;
        if(hex_len > byte_len)
            throw std::domain_error("Integer too large to be represented by "
                                    + std::to_string(byte_len) + " hex characters!");
        std::vector<char> hex(2*byte_len, '0');

        auto [ignored, ec] = std::to_chars(hex.data() + hex.size() - hex_len,
                                           hex.data() + hex.size(), N, 16);
        if(ec != std::errc())
            throw std::runtime_error("Could not convert integer to hex "
                                     "representation.");

        std::vector<unsigned char> result(byte_len);
        sodium_hex2bin(result.data(), byte_len, hex.data(), hex.size(),
                       NULL, NULL, NULL);

        return result;
    }

    /**
     * \brief Shortcut for binaryRepresentation() with \a byte_len =
     *        \b crypto_stream_chacha20_ietf_NONCEBYTES.
     *
     * Used to convert a password counter to hex representation prior to
     * password derivation.
     *
     * \sa binaryRepresentation() for details and derivePasswordBlock() for a
     *     use case
     */
    std::vector<unsigned char> counter2nonce(uint64_t counter) {
        return binaryRepresentation(counter,
                                    crypto_stream_chacha20_ietf_NONCEBYTES);
    }

    /**
     * \brief Computes \"N \b mod base\", where N is the integer representation
     *        of the data \a block.
     *
     * This function is used to obtain a pseudo random character from a
     * deterministically derived block of bytes. For the interpretation of
     * \a block as an integer, see the documentation of asciiBitString().
     *
     * \param [in] block data that is converted to an integer using
     *             asciiBitString(), which is then computed modulo \a base
     * \param [in] base the divisor of the modulo operation. Needs to be >= 2.
     * \return \b int( \a block ) \b mod \a base
     */
    unsigned int block_modulo(const UCV& block, unsigned int base) {

        UCV_uptr bits = asciiBitString(block);

        std::vector<unsigned int> R(1, 1); // [2**i % base for i in range(512)]
        std::set<unsigned int> S {1}; // use set for time efficient searches

        unsigned int next = 2 % base;
        while(S.find(next) == S.end()) { // push back numbers (2^i % base) until
            R.push_back(next);           // the numbers repeat
            S.insert(next);
            next = (2*R.back()) % base;
        }
        // index of first element that repeats:
        unsigned int p = find(R.begin(), R.end(), next) - R.begin();
        // elements with index < p occur only once, while the remaining ones
        // repeat indefinitely. This allows to efficiently compute (2^i % base)
        // for any i:
        // 2**i % base = R[i] if i<R.size else R[p + (i-p) % (R.size()-p)]

        unsigned int result = 0;
        for(size_t i=0; i<bits->size(); ++i) {
            if((*bits)[i] == '1') {
                unsigned int e = bits->size() - 1 - i; // exponent 2^e
                if(e < R.size()) result += R[e];
                else result += R[p+(e-p)%(R.size()-p)];
            }
        }

        return result % base;
    }

    /**
     * \brief Computes a pseudo random block of data using a CSPRNG given a
     *        \a seed and \a counter
     *
     * \warning The strength of the returned password depends on how
     *          unpredictable the provided \a seed is. The caller has to
     *          guarantee, that the \a seed has sufficient entropy.
     *
     * Provided with a certain set of parameters, this function will always
     * produce the same output. It can be used to deterministically derive
     * passwords. Random passwords can be generated by using a random \a seed.     *
     * The result is a sequence of util::PW_BLOCK_SIZE pseudo random bytes. It
     * can be used to derive a single character for a password by computing its
     * value modulo size(alphabet) using block_modulo().
     *
     * \note that the maximum number of blocks that can be derived is
     *       \f$2^{32}$. If each block is converted to a single ASCII character
     *       (which has the size of one byte), the longest possible password's
     *       size is \f$2^{32} = 4\;\textrm{GiB}$.
     *
     * The length of the provided \a seed has to be util::PW_SEED_LENGTH. This
     * has to be guaranteed by the caller. If the size of the \a seed is less
     * than that, the behaviour is not defined; if it is larger, the excess
     * bytes will be ignored. In order to not compute the nonce for the ChaCha20
     * algorithm more than once, it has to be provided by the callee. The nonce
     * can be obtained using util::counter2nonce(counter).
     *
     * The current implementation uses the ChaCha20 stream cipher to obtain a
     * pseudo random block by encrypting a block of zeros. The user provided
     * counter needs to be converted using counter2nonce() to a platform
     * independent representation and used is used as a nonce. It allows to
     * generate new passwords by just increasing the counter without changing
     * the seed.
     *
     * \note For efficiency, the UCV \a out is not initialized on
     *       every call but needs to be initialized by the caller. It has to be
     *       large enough to store at least util::PW_BLOCK_SIZE bytes. If its
     *       size is larger than that, the bytes beginning from
     *       util::PW_BLOCK_SIZE+1 will be ignored.
     *
     * \param [in] seed cryptographically secure data of length
     *             PW_SEED_LENGTH. The caller needs to ensure, that the provided
     *             seed has the expected size.
     * \param [in] nonce a nonce that should be created using counter2nonce().
     *             Different values will produce different passwords.
     * \param [in] block_counter index of the password block that is to be
     *             computed
     * \param [out] out UCV in which the derived password block
     *              will be stored. Needs to be pre-initialized with size of at
     *              least util::PW_BLOCK_SIZE bytes.
     */
    void derivePasswordBlock(const UCV& seed,
                             std::vector<unsigned char> const& nonce,
                             uint32_t block_counter,
                             UCV& out) {

        int status = crypto_stream_chacha20_ietf_xor_ic
                        (out.data(), chacha_block_zeros, PW_BLOCK_SIZE,
                         nonce.data(), block_counter, seed.data());
        if(status != 0)
            throw std::runtime_error("Failed to compute passphrase block using "
                                     "'crypto_stream_chacha20_ietf_xor_ic'.");
    }

} // namespace sinaps::SecureMemory::util
