// Copyright 2023 xwst <xwst@gmx.net>

/*
This file is part of sinaps.

sinaps is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

sinaps is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with sinaps. If not, see <http://www.gnu.org/licenses/>.
*/

#include "secure_memory/vector.h"

#include <sodium.h>

/**
 * \namespace sinaps::SecureMemory
 * \brief Namespace for classes, containers, related functions, etc., that store
 *        sensitive data in memory allocated by sodium_malloc().
 * \cond
 *      Without namespace documentation, the documentation of the typedefs is
 *      not generated
 * \endcond
 */

/**
 * \typedef SecureMemory::Vector
 * \brief Template for a STL vector that allocates memory using 'sodium_malloc'.
 *
 * \sa the specialization SecureMemory::UCharVector
 */

/**
 * \typedef SecureMemory::UCV
 * \brief Specialization of SecureMemory::Vector for unsigned characters
 *
 * If possible, i.e. when the underlying data does not need to be shared, it is
 * advisable to use SecureMemory::uptrUCV instead.
 */

/**
 * \typedef SecureMemory::const_UCV
 * \brief Typedef to a constant SecureMemory::UCV.
 */

/**
 * \typedef SecureMemory::uptrUCV
 * \brief STL unique pointer to a SecureMemory::UCV.
 *
 * This typedef is used to make sure ownership of secret data is not shared and
 * the secret data is not copied by accident.
 */

namespace sinaps::SecureMemory {

    /**
     * \name SecureMemory::Vector - Access Restrictions
     * \brief Auxiliary functions that restrict access to the UCharVector
     *        managed by a unique pointer.
     *
     * See the documentation of [libsodium]
     * (https://doc.libsodium.org/memory_management) for details.
     */
    /**@{*/

    /** Prevents read and write access. */
    template <typename T>
    void setNoAccess(std::unique_ptr<Vector<T>>& ptr_vec) {
        sodium_mprotect_noaccess(ptr_vec->data());
    }

    /** Makes the data read only. */
    template <typename T>
    void setReadOnly(std::unique_ptr<Vector<T>>& ptr_vec) {
        sodium_mprotect_readonly(ptr_vec->data());
    }

    /** Allows read and write access. */
    template <typename T>
    void setReadWrite(std::unique_ptr<Vector<T>>& ptr_vec) {
        sodium_mprotect_readwrite(ptr_vec->data());
    }

    /**@}*/

} // namespace sinaps::SecureMemory
