// Copyright 2023 xwst <xwst@gmx.net>

/*
This file is part of sinaps.

sinaps is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

sinaps is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with sinaps. If not, see <http://www.gnu.org/licenses/>.
*/

#include "unittest/helpers.h"

#include <cstdio>
#include <cstring>

#include <glibmm.h>

#include "secure_memory/passphrase.h"
#include "secure_memory/vector.h"

/**
 * \namespace sinaps::UnitTests
 * \brief Collection of classes and utility functions that form the testing
 *        suite.
 */

/**
 * \class UnitTests::RandomParameters
 * \brief Class that encapsulates the parameters needed for password derivation
 *        and creates random parameters on construction.
 */

/**
 * \var RandomParameters::pin_len
 * \brief Length of \link RandomParameters::pin pin \endlink
 */
/**
 * \var RandomParameters::context_len
 * \brief Length of \link RandomParameters::context context. \endlink
 */
/**
 * \var RandomParameters::alphabet_len
 * \brief Length of \link RandomParameters::alphabet alphabet. \endlink
 */
/**
 * \var RandomParameters::pin
 * \brief See SecureMemory::Passphrase::Passphrase<!--
 *                                   -->(const_UCV&, unsigned long long, size_t)
 */
/**
 * \var RandomParameters::context
 * \brief See SecureMemory::Passphrase::custom() for details.
 */
/**
 * \var RandomParameters::alphabet
 * \brief Internal variable that stores the set of characters that is used as an
 * alphabet in any custom encoding.
 *
 * \sa SecureMemory::Passphrase::custom() for details.
 */
/**
 * \var RandomParameters::encoding
 * \brief Stores either a specific encoding like 'hex' or 'base64' or a set of
 * characters that represents an alphabet for custom encoding.
 *
 * \sa RandomParameters::alphabet
 */
/**
 * \var RandomParameters::counter
 * \brief See SecureMemory::Passphrase::custom() for details.
 */
/**
 * \var RandomParameters::opslimit
 * \brief See SecureMemory::Passphrase::Passphrase<!--
 *                                   -->(const_UCV&, unsigned long long, size_t)
 */
/**
 * \var RandomParameters::memlimit
 * \brief See SecureMemory::Passphrase::Passphrase<!--
 *                                   -->(const_UCV&, unsigned long long, size_t)
 */
/**
 * \var RandomParameters::length
 * \brief See SecureMemory::Passphrase::custom() for details.
 */

namespace sinaps::UnitTests {

    /**
     * \brief Constructs a 'RandomParameter' object and picks all member
     *        variables at random.
     */
    RandomParameters::RandomParameters() {

        this->pin = std::make_unique<SecureMemory::Vector<char>>();
        this->pin->resize(random(2, 50));
        this->context.resize(random(2, 100));
        randomString(*pin);
        randomString(context);

        if(random(0, 1)) {                         // Fill vector with
            this->encoding.resize(random(10, 50)); // characters that will be
            randomString(encoding);                // used for encoding
        } // If the vector remains empty, the raw password stream will be tested

        this->opslimit = random(1, 4);
        this->memlimit = random(8, 262144); // used memory in bytes/1024
        this->counter = std::rand();
        this->length = random(2, 60);
    }

    /**
     * \brief Auxiliary function that returns a random number between \a min and
     *        \a max.
     *
     * The boundaries \a min and \a max are included and might be returned.
     *
     * \param [in] min the minimum number that may be returned
     * \param [in] min the maximum number that may be returned
     */
    size_t RandomParameters::random(size_t min, size_t max) {
        return min + std::rand() * (max - min) / RAND_MAX;
    }

    /**
     * \brief Auxiliary function that fills a given STL vector<char> with random
     *        printable ASCII characters
     *
     * The variable \a vec is assumed to be a STL vector of type 'char'. It will
     * be overwritten by random data that is not null-terminated. Its size is
     * not changed.
     *
     * The set of ASCII characters that may appear in \a vec are the printable
     * ASCII characters with decimal representations 32 to 126 (both included).
     *
     * \param [in,out] vec the vector in which the randomly generated string
     *                     will be stored.
     */
    template <typename V>
    void RandomParameters::randomString(V& vector) {
        int minASCII =  32;
        int maxASCII = 126;

        for(unsigned int i=0; i<vector.size(); ++i)
            vector[i] = static_cast<char>(random(minASCII, maxASCII));
    }


    /**
     * \brief Auxiliary function to obtain a derived password from the reference
     *        implementation.
     *
     * For a detailed description of the parameters, see the documentation of
     * the functions
     * SecureMemory::Passphrase::Passphrase<!--
     *                               -->(const_UCV&, unsigned long long, size_t)
     * and SecureMemory::Passphrase::custom().
     *
     * \sa cpp_implementation()
     */
    std::string referenceImplementation(RandomParameters& p) {

        std::vector<std::string> argv{REFERENCE_IMPL_CMD};
        int in, out;
        gsize bytes_written;
        Glib::spawn_async_with_pipes("", argv, Glib::SpawnFlags::DEFAULT, {},
                                    nullptr, &in, &out, nullptr);
        auto ic = Glib::IOChannel::create_from_fd(in);
        auto oc = Glib::IOChannel::create_from_fd(out);
        ic->write(p.pin->data(), p.pin->size(), bytes_written);
        ic->write(Glib::ustring::compose("\n%1\n%2\n", p.opslimit, p.memlimit));
        ic->write(p.context.data(), p.context.size(), bytes_written);
        ic->write(Glib::ustring::compose("\n%1\n", p.counter));


        if(p.encoding.size() > 0)
            ic->write(p.encoding.data(), p.encoding.size(), bytes_written);
        else ic->write("raw");

        ic->write(Glib::ustring::compose("\n%1\n", p.length));
        ic->close();

        std::vector<char> output;
        output.resize(p.length);
        oc->set_encoding("");
        oc->read(output.data(), p.length, bytes_written);
        oc->close();
        return std::string(output.data(), bytes_written);
    }

    /**
     * \brief Creates a STL vector<char> from a C-style string.
     *
     * \param [in] str a C-style string. Needs to be null-terminated.
     * \return a std::vector<char> representation of the C-string without
     *         null-termination.
     */
    template <typename Allocator>
    std::vector<char, Allocator> CString2Vector(const char* str) {
        int length = strlen(str);
        std::vector<char, Allocator> result;
        result.reserve(length);
        for(int i=0; i<length; ++i)
            result.push_back(str[i]);
        return result;
    }


    /**
     * \brief Auxiliary function to obtain a derived password from the C++
     *        implementation.
     *
     * For a detailed description of the parameters, see the documentation of
     * the functions
     * SecureMemory::Passphrase::Passphrase<!--
     *                              -->(const_UCV&, unsigned long long, size_t),
     * SecureMemory::Passphrase::custom() and
     * SecureMemory::Passphrase::writeRaw().
     *
     * \warning This function invalidates p.pin, as it std::moves it into
     *          SecureMemory::Passphrase.
     *
     * \sa referenceImplementation()
     */
    std::string cpp_implementation(RandomParameters& p) {

        SecureMemory::Passphrase<char> M
            = SecureMemory::Passphrase<char>(std::move(p.pin), p.opslimit,
                                               1024*p.memlimit);

        if(p.encoding.size() > 0) {
            sinaps::SecureMemory::Vector_uptr<char> result;
            result = M.custom(p.context, p.counter, p.length, p.encoding);
            return std::string(result->data(), result->size());
        } else {
            std::ostringstream result;
            M.writeRaw(result, p.context, p.counter, p.length);
            return result.str();
        }
    }


    /**
     * \brief Create passwords with random parameters and verify that the
     *        reference implementation and C++ code agree.
     *
     * The parameters are obtained by instantiating a RandomPassword object. The
     * number of passwords to be checked can is controlled by the
     * \a numberOfTests parameter.
     *
     * \param [in] numberOfTests how many times a set of random parameters is
     *                           created and the corresponding password is
     *                           verified using the reference implementation.
     */
    void randomPasswordDerivationTests(int numberOfTests = 20) {

        for(int i=0; i<numberOfTests; ++i) {
            RandomParameters RP;

            char parameters [1024];
            RP.pin->push_back('\0');
            RP.context.push_back('\0');
            RP.encoding.push_back('\0');
            std::snprintf(parameters, 1023, "%s, %i, %i, %s, %li, %s, %i",
                          RP.pin->data(), RP.opslimit, RP.memlimit,
                          RP.context.data(), RP.counter, RP.encoding.data(),
                          RP.length);
            RP.pin->pop_back();
            RP.context.pop_back();
            RP.encoding.pop_back();

            std::string A = referenceImplementation(RP);
            std::string B = cpp_implementation(RP);

            TSM_ASSERT_EQUALS(parameters, A.size(), B.size());
            TSM_ASSERT_SAME_DATA(parameters, A.c_str(), B.c_str(), A.size());
        }
    }


    /**
     * \brief Verify that derived passwords of different length (but same
     *        parameters otherwise) are identical when removing the excess bytes
     *        of the longer password.
     *
     * Two passwords generated with the same parameters but lenght m and n, m<n,
     * should be identical on the first m bytes. This function loops from
     * length 1 to \a maxLength and verifies that increasing the length by one
     * only appends a new character but does not change the previous ones. This
     * is repeated for \a cycles passwords.
     *
     * \param [in] maxLength maximum length for which the above property is
     *                       checked
     * \param [in] cycles number of different passwords for which the property
     *             is verified
     */
    void passwordAsSequenceTests(int maxLength = 100, int cycles = 5) {

        for(int c=0; c<cycles; ++c) {
            RandomParameters RP;

            sinaps::SecureMemory::Passphrase M
                = SecureMemory::Passphrase<char>(std::move(RP.pin), RP.opslimit,
                                                 1024*RP.memlimit);

            if(RP.encoding.size() > 0) {
                sinaps::SecureMemory::Vector_uptr<char> old
                    = M.custom(RP.context, RP.counter, 1, RP.encoding, 0);

                for(int L=2; L<=maxLength; ++L) {
                    sinaps::SecureMemory::Vector_uptr<char> pw
                        = M.custom(RP.context, RP.counter, L, RP.encoding, 0);

                    TS_ASSERT_SAME_DATA(pw->data(), old->data(), old->size());
                    old = std::move(pw);
                }
            } else {
                std::ostringstream old_oss;
                M.writeRaw(old_oss, RP.context, RP.counter, 1, 0);
                std::string old = old_oss.str();

                for(int L=2; L<=maxLength; ++L) {
                    std::ostringstream pw_oss;
                    M.writeRaw(pw_oss, RP.context, RP.counter, L, 0);
                    std::string pw = pw_oss.str();
                    TS_ASSERT_EQUALS(pw.size(), L);
                    TS_ASSERT(pw.starts_with(old));
                    old = std::move(pw);
                }
            }
        }
    }
} // namespace sinaps::UnitTests
