#!/usr/bin/python3

# Copyright 2023 xwst <xwst@gmx.net>


# This file is part of sinaps.
#
# sinaps is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# sinaps is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with sinaps. If not, see <http://www.gnu.org/licenses/>.


from math import ceil, log2
from binascii import hexlify, unhexlify, b2a_base64
from argon2.low_level import hash_secret_raw as argon2, Type as argon2_type
from hashlib import blake2b
from Cryptodome.Cipher import ChaCha20
import sys

# Constants
salt_str       = "sinaps is not a password store".encode("ascii")
salt_len       = 16 # byte = crypto_pwhash_SALTBYTES
key_len        = 32 # byte
argon2_threads =  1 # libsodium only supports one thread
argon2_version = 19 # = 0x13 (equals the version named 1.3 in libsodium)
nonce_width    = 12 # fixed width of nonce in ChaCha20_IETF (can represent at
                    # most 2**48-1 = "f"*12 in ascii-hex encoding)
counter_max    = 16**(2*nonce_width) - 1

# User input
pin      = input().encode("ascii") # user pin that is used to derive passwords
opslimit = int(input()) # number of argon2 iterations
memlimit = int(input()) # memory utilized by argon2
context  = input() # context/salt for password derivation
counter  = input() # password counter
counter  = int(counter) if counter else 0
assert(counter <= counter_max and counter >= 0)
encoding = input() # encoding; options are 'raw', 'bin', 'dec', 'hex' and
                   # 'base64'. If none of these classifiers is specified, the
                   # provided characters will be used for the encoding
length = int(input("")) # length the calculated password shall have
assert(length > 0)


# Computation
## Salt
h = blake2b(digest_size = salt_len)
h.update(salt_str)
salt = h.digest()

## Passphrase
passphrase = argon2(pin, salt, opslimit, memlimit, argon2_threads, key_len,
                   argon2_type.ID, argon2_version)

## Derive Key Seed
h = blake2b(key = passphrase, digest_size = key_len)
h.update(context.encode("ascii"))
seed = h.digest()

## Derive Key
# Standard encoding: convert one 512 bit block to a single character
# Raw: encode raw data using standard algorithms
blocksize = 64
bit_per_character = 6 if encoding == "base64" else \
                    4 if encoding == "hex"    else \
                    8*blocksize   if encoding == "dec"    else \
                    1 if encoding == "bin"    else \
                    8 if encoding == "raw"    else \
                    8*blocksize

nonce = unhexlify("{0:0{1}x}".format(counter, 2*nonce_width))
plain = b'\0' * ceil(length * bit_per_character / 8)
DPRNG = ChaCha20.new(key=seed, nonce=nonce)
dkey = DPRNG.encrypt(plain)

def convert_block(block, alphabet):
    int_data = int(hexlify(block), base=16)
    return alphabet[int_data % len(alphabet)]

## Encoding & Output
def base_conversion(data, alphabet):
    alphabet = tuple(sorted(set(alphabet))) # sanitize alphabet
    result = ""
    while len(data) > 0:
        result += convert_block(data[0:blocksize], alphabet)
        data = data[blocksize:]
    return result

if encoding == "hex":
    dkeyenc = hexlify(dkey).decode("ascii")[:length]
elif encoding == "base64":
    dkeyenc = b2a_base64(dkey).decode("ascii")
    dkeyenc = dkeyenc[:-1].replace("=", "") # Remove newline-char and padding
    dkeyenc = dkeyenc[:length]
elif encoding == "dec":
    dkeyenc = base_conversion(dkey, "0123456879")
elif encoding == "bin":
    dkeyenc = format(int(hexlify(dkey), 16), 'b')[:length]
elif encoding == "raw":
    dkeyenc = dkey
else:
    dkeyenc = base_conversion(dkey, encoding)

try:
    sys.stdout.buffer.write(dkeyenc)
except:
    print(dkeyenc)
